import asyncio
import mimetypes
from collections.abc import AsyncGenerator, AsyncIterator, Callable
from pathlib import Path

from notionary.exceptions.file_upload import UploadFailedError, UploadTimeoutError
from notionary.file_upload.client import FileUploadHttpClient
from notionary.file_upload.config import NOTION_SINGLE_PART_MAX_SIZE, FileUploadConfig
from notionary.file_upload.file.reader import FileContentReader
from notionary.file_upload.query import FileUploadQuery, FileUploadQueryBuilder
from notionary.file_upload.schemas import FileUploadResponse, FileUploadStatus
from notionary.file_upload.validation.factory import (
    create_bytes_upload_validation_service,
    create_file_upload_validation_service,
)
from notionary.utils.mixins.logging import LoggingMixin


class NotionFileUpload(LoggingMixin):
    def __init__(
        self,
        client: FileUploadHttpClient | None = None,
        config: FileUploadConfig | None = None,
        file_reader: FileContentReader | None = None,
    ):
        self._client = client or FileUploadHttpClient()
        self._config = config or FileUploadConfig()
        self._file_reader = file_reader or FileContentReader(config=self._config)

    async def upload_file(
        self, file_path: Path, filename: str | None = None
    ) -> FileUploadResponse:
        file_path = Path(file_path)

        if not file_path.is_absolute() and self._config.base_upload_path:
            file_path = Path(self._config.base_upload_path) / file_path
        file_path = file_path.resolve()

        validator_service = create_file_upload_validation_service(file_path)
        await validator_service.validate()

        file_size = file_path.stat().st_size
        filename = filename or file_path.name
        content_type = self._guess_content_type(filename)

        if self._fits_in_single_part(file_size):
            content = await self._file_reader.read_full_file(file_path)
            return await self._upload_single_part_content(
                content, filename, content_type
            )
        else:
            return await self._upload_multi_part_content(
                filename,
                content_type,
                file_size,
                self._file_reader.read_file_chunks(file_path),
            )

    async def upload_from_bytes(
        self,
        file_content: bytes,
        filename: str,
        content_type: str | None = None,
    ) -> FileUploadResponse:
        file_size = len(file_content)

        validator_service = create_bytes_upload_validation_service(
            filename=filename,
            file_size_bytes=file_size,
        )
        await validator_service.validate()

        content_type = content_type or self._guess_content_type(filename)

        if self._fits_in_single_part(file_size):
            return await self._upload_single_part_content(
                file_content, filename, content_type
            )

        return await self._upload_multi_part_content(
            filename,
            content_type,
            file_size,
            self._file_reader.bytes_to_chunks(file_content),
        )

    async def _upload_single_part_content(
        self, content: bytes, filename: str, content_type: str | None
    ) -> FileUploadResponse:
        file_upload = await self._client.create_single_part_upload(
            filename=filename,
            content_type=content_type,
        )

        await self._client.send_file_content(
            file_upload_id=file_upload.id,
            file_content=content,
            filename=filename,
        )

        self.logger.info(
            "Single-part content sent, waiting for completion... (ID: %s)",
            file_upload.id,
        )
        return await self._wait_for_completion(file_upload.id)

    async def _upload_multi_part_content(
        self,
        filename: str,
        content_type: str | None,
        file_size: int,
        chunk_generator: AsyncGenerator[bytes],
    ) -> FileUploadResponse:
        part_count = self._calculate_part_count(file_size)

        file_upload = await self._client.create_multi_part_upload(
            filename=filename,
            content_type=content_type,
            number_of_parts=part_count,
        )

        await self._send_parts(file_upload.id, filename, part_count, chunk_generator)

        await self._client.complete_upload(file_upload.id)

        self.logger.info(
            "Multi-part content sent, waiting for completion... (ID: %s)",
            file_upload.id,
        )
        return await self._wait_for_completion(file_upload.id)

    async def _send_parts(
        self,
        file_upload_id: str,
        filename: str,
        total_parts: int,
        chunk_generator: AsyncGenerator[bytes],
    ) -> None:
        part_number = 1
        try:
            async for chunk in chunk_generator:
                await self._client.send_file_content(
                    file_upload_id=file_upload_id,
                    file_content=chunk,
                    filename=filename,
                    part_number=part_number,
                )

                self.logger.debug("Uploaded part %d/%d", part_number, total_parts)
                part_number += 1

        except Exception as e:
            raise UploadFailedError(
                file_upload_id=file_upload_id,
                reason=f"Failed to upload part {part_number}/{total_parts}: {e}",
            ) from e

    def _fits_in_single_part(self, file_size: int) -> bool:
        return file_size <= NOTION_SINGLE_PART_MAX_SIZE

    def _guess_content_type(self, filename: str) -> str | None:
        content_type, _ = mimetypes.guess_type(filename)
        return content_type

    def _calculate_part_count(self, file_size: int) -> int:
        return (
            file_size + self._config.multi_part_chunk_size - 1
        ) // self._config.multi_part_chunk_size

    async def get_upload_status(self, file_upload_id: str) -> str:
        try:
            upload_info = await self._client.get_file_upload(file_upload_id)
            return upload_info.status
        except Exception as e:
            raise UploadFailedError(file_upload_id, reason=str(e)) from e

    async def _wait_for_completion(
        self,
        file_upload_id: str,
        timeout_seconds: int | None = None,
    ) -> FileUploadResponse:
        timeout = timeout_seconds or self._config.max_upload_timeout

        try:
            return await asyncio.wait_for(
                self._poll_status_until_complete(file_upload_id), timeout=timeout
            )

        except TimeoutError as e:
            raise UploadTimeoutError(file_upload_id, timeout) from e

    async def _poll_status_until_complete(
        self, file_upload_id: str
    ) -> FileUploadResponse:
        while True:
            upload_info = await self._client.get_file_upload(file_upload_id)

            if upload_info.status == FileUploadStatus.UPLOADED:
                self.logger.info("Upload completed: %s", file_upload_id)
                return upload_info

            if upload_info.status == FileUploadStatus.FAILED:
                raise UploadFailedError(file_upload_id)

            await asyncio.sleep(self._config.poll_interval)

    async def get_uploads(
        self,
        *,
        filter_fn: Callable[[FileUploadQueryBuilder], FileUploadQueryBuilder]
        | None = None,
        query: FileUploadQuery | None = None,
    ) -> list[FileUploadResponse]:
        resolved_query = self._resolve_query(filter_fn=filter_fn, query=query)
        return await self._client.list_file_uploads(query=resolved_query)

    async def iter_uploads(
        self,
        *,
        filter_fn: Callable[[FileUploadQueryBuilder], FileUploadQueryBuilder]
        | None = None,
        query: FileUploadQuery | None = None,
    ) -> AsyncIterator[FileUploadResponse]:
        resolved_query = self._resolve_query(filter_fn=filter_fn, query=query)
        async for upload in self._client.list_file_uploads_stream(query=resolved_query):
            yield upload

    def _resolve_query(
        self,
        filter_fn: Callable[[FileUploadQueryBuilder], FileUploadQueryBuilder]
        | None = None,
        query: FileUploadQuery | None = None,
    ) -> FileUploadQuery:
        if filter_fn and query:
            raise ValueError("Use either filter_fn OR query, not both")

        if filter_fn:
            builder = FileUploadQueryBuilder()
            configured_builder = filter_fn(builder)
            return configured_builder.build()

        return query or FileUploadQuery()



import asyncio

from notionary.blocks.rich_text.rich_text_markdown_converter import (
    RichTextToMarkdownConverter,
)
from notionary.comments.models import Comment
from notionary.comments.schemas import CommentDto
from notionary.user.base import BaseUser
from notionary.user.client import UserHttpClient
from notionary.utils.mixins.logging import LoggingMixin


class CommentFactory(LoggingMixin):
    UNKNOWN_AUTHOR = "Unknown Author"

    def __init__(
        self,
        http_client: UserHttpClient | None = None,
        markdown_converter: RichTextToMarkdownConverter | None = None,
    ) -> None:
        self.http_client = http_client
        self.markdown_converter = markdown_converter or RichTextToMarkdownConverter()

    async def create_from_dto(self, dto: CommentDto) -> Comment:
        author_name, content = await asyncio.gather(
            self._resolve_user_name(dto), self._resolve_content(dto)
        )

        return Comment(author_name=author_name, content=content)

    async def _resolve_user_name(self, dto: CommentDto) -> str:
        created_by_id = dto.created_by.id

        try:
            return await BaseUser.from_id_auto(created_by_id, self.http_client)
        except Exception:
            self.logger.warning(
                f"Failed to resolve user name for user_id: {created_by_id}",
                exc_info=True,
            )

        return self.UNKNOWN_AUTHOR

    async def _resolve_content(self, dto: CommentDto) -> str:
        return await self.markdown_converter.to_markdown(dto.rich_text)



from typing import override

from notionary.blocks.rich_text.rich_text_markdown_converter import (
    RichTextToMarkdownConverter,
)
from notionary.blocks.schemas import Block, BlockType
from notionary.page.content.renderer.context import MarkdownRenderingContext
from notionary.page.content.renderer.renderers.base import BlockRenderer
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class CalloutRenderer(BlockRenderer):
    def __init__(
        self,
        syntax_registry: SyntaxDefinitionRegistry | None = None,
        rich_text_markdown_converter: RichTextToMarkdownConverter | None = None,
    ) -> None:
        super().__init__(syntax_registry=syntax_registry)
        self._rich_text_markdown_converter = (
            rich_text_markdown_converter or RichTextToMarkdownConverter()
        )

    @override
    def _can_handle(self, block: Block) -> bool:
        return block.type == BlockType.CALLOUT

    @override
    async def _process(self, context: MarkdownRenderingContext) -> None:
        content = await self._extract_callout_content(context.block)

        if not content:
            context.markdown_result = ""
            return

        icon = await self._extract_callout_icon(context.block)

        callout_start_delimiter = (
            self._syntax_registry.get_callout_syntax().start_delimiter
        )

        result = (
            f'{callout_start_delimiter}({content} "{icon}")'
            if icon
            else f"{callout_start_delimiter}({content})"
        )

        if context.indent_level > 0:
            result = context.indent_text(result)

        context.markdown_result = result

    async def _extract_callout_icon(self, block: Block) -> str:
        if not block.callout or not block.callout.icon:
            return ""
        return block.callout.icon.emoji or ""

    async def _extract_callout_content(self, block: Block) -> str:
        if not block.callout or not block.callout.rich_text:
            return ""
        return await self._rich_text_markdown_converter.to_markdown(
            block.callout.rich_text
        )



import difflib
from collections.abc import Callable
from dataclasses import dataclass
from typing import Generic, TypeVar

T = TypeVar("T")


@dataclass(frozen=True)
class _MatchResult(Generic[T]):
    item: T
    similarity: float


def find_best_match(
    query: str,
    items: list[T],
    text_extractor: Callable[[T], str],
    min_similarity: float,
) -> T | None:
    matches = _find_best_matches(query, items, text_extractor, min_similarity, limit=1)
    return matches[0].item if matches else None


def find_all_matches(
    query: str,
    items: list[T],
    text_extractor: Callable[[T], str],
    min_similarity: float,
) -> list[T]:
    matches = _find_best_matches(
        query, items, text_extractor, min_similarity, limit=None
    )
    return [match.item for match in matches]


def _find_best_matches(
    query: str,
    items: list[T],
    text_extractor: Callable[[T], str],
    min_similarity: float = 0.0,
    limit: int | None = None,
) -> list[_MatchResult[T]]:
    results = []

    for item in items:
        text = text_extractor(item)
        similarity = _calculate_similarity(query, text)

        if similarity >= min_similarity:
            results.append(_MatchResult(item=item, similarity=similarity))

    results = _sort_by_highest_similarity_first(results)

    if limit:
        return results[:limit]

    return results


def _sort_by_highest_similarity_first(
    results: list[_MatchResult],
) -> list[_MatchResult]:
    return sorted(results, key=lambda x: x.similarity, reverse=True)


def _calculate_similarity(query: str, target: str) -> float:
    return difflib.SequenceMatcher(
        isjunk=None,
        a=query.lower().strip(),
        b=target.lower().strip(),
    ).ratio()



import asyncio
import os

import httpx
from dotenv import load_dotenv

from notionary.exceptions.api import (
    NotionApiError,
    NotionAuthenticationError,
    NotionConnectionError,
    NotionPermissionError,
    NotionRateLimitError,
    NotionResourceNotFoundError,
    NotionServerError,
    NotionValidationError,
)
from notionary.http.models import HttpMethod
from notionary.shared.typings import JsonDict
from notionary.utils.mixins.logging import LoggingMixin

load_dotenv(override=True)


class NotionHttpClient(LoggingMixin):
    BASE_URL = "https://api.notion.com/v1"
    NOTION_VERSION = "2025-09-03"

    def __init__(self, timeout: int = 30):
        self.token = self._find_token()
        if not self.token:
            raise ValueError(
                "No Notion API token found in environment variables. Please set one of these environment variables: "
                "NOTION_SECRET, NOTION_INTEGRATION_KEY, or NOTION_TOKEN"
            )

        self.headers = {
            "Authorization": f"Bearer {self.token}",
            "Content-Type": "application/json",
            "Notion-Version": self.NOTION_VERSION,
        }

        self.client: httpx.AsyncClient | None = None
        self.timeout = timeout
        self._is_initialized = False

    def __del__(self):
        if not hasattr(self, "client") or not self.client:
            return

        try:
            loop = asyncio.get_event_loop()
            if not loop.is_running():
                self.logger.warning(
                    "Event loop not running, could not auto-close NotionHttpClient"
                )
                return

            loop.create_task(self.close())
            self.logger.debug("Created cleanup task for NotionHttpClient")
        except RuntimeError:
            self.logger.warning(
                "No event loop available for auto-closing NotionHttpClient"
            )

    async def __aenter__(self):
        await self._ensure_initialized()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self) -> None:
        if not hasattr(self, "client") or not self.client:
            return

        await self.client.aclose()
        self.client = None
        self._is_initialized = False
        self.logger.debug("NotionHttpClient closed")

    async def get(
        self, endpoint: str, params: JsonDict | None = None
    ) -> JsonDict | None:
        return await self._make_request(HttpMethod.GET, endpoint, params=params)

    async def post(
        self, endpoint: str, data: JsonDict | None = None
    ) -> JsonDict | None:
        return await self._make_request(HttpMethod.POST, endpoint, data)

    async def patch(
        self, endpoint: str, data: JsonDict | None = None
    ) -> JsonDict | None:
        return await self._make_request(HttpMethod.PATCH, endpoint, data)

    async def delete(self, endpoint: str) -> JsonDict | None:
        return await self._make_request(HttpMethod.DELETE, endpoint)

    async def _make_request(
        self,
        method: HttpMethod,
        endpoint: str,
        data: JsonDict | None = None,
        params: JsonDict | None = None,
    ) -> JsonDict | None:
        await self._ensure_initialized()

        url = f"{self.BASE_URL}/{endpoint.lstrip('/')}"
        try:
            self.logger.debug("Sending %s request to %s", method.value.upper(), url)

            request_kwargs = {}

            # Add query parameters for GET requests
            if params:
                request_kwargs["params"] = params

            if (
                method.value in [HttpMethod.POST.value, HttpMethod.PATCH.value]
                and data is not None
            ):
                request_kwargs["json"] = data

            response: httpx.Response = await getattr(self.client, method.value)(
                url, **request_kwargs
            )

            response.raise_for_status()
            result_data = response.json()
            self.logger.debug("Request successful: %s", url)
            return result_data

        except httpx.HTTPStatusError as e:
            self._handle_http_status_error(e)
        except httpx.RequestError as e:
            raise NotionConnectionError(
                f"Failed to connect to Notion API: {e!s}. Please check your internet connection and try again."
            ) from e

    def _handle_http_status_error(self, e: httpx.HTTPStatusError) -> None:
        status_code = e.response.status_code
        response_text = e.response.text

        if status_code == 401:
            raise NotionAuthenticationError(
                "Invalid or missing API key. Please check your Notion integration token.",
                status_code=status_code,
                response_text=response_text,
            )
        if status_code == 403:
            raise NotionPermissionError(
                "Insufficient permissions. Please check your integration settings at "
                "https://www.notion.so/profile/integrations and ensure the integration "
                "has access to the required pages/databases.",
                status_code=status_code,
                response_text=response_text,
            )
        if status_code == 404:
            raise NotionResourceNotFoundError(
                "The requested resource was not found. Please verify the page/database/datasource ID.",
                status_code=status_code,
                response_text=response_text,
            )
        if status_code == 400:
            raise NotionValidationError(
                f"Invalid request data. Please check your input parameters: {response_text}",
                status_code=status_code,
                response_text=response_text,
            )
        if status_code == 429:
            raise NotionRateLimitError(
                "Rate limit exceeded. Please wait before making more requests.",
                status_code=status_code,
                response_text=response_text,
            )
        if 500 <= status_code < 600:
            raise NotionServerError(
                "Notion API server error. Please try again later.",
                status_code=status_code,
                response_text=response_text,
            )

        raise NotionApiError(
            f"API request failed with status {status_code}: {response_text}",
            status_code=status_code,
            response_text=response_text,
        )

    def _find_token(self) -> str | None:
        token = next(
            (
                os.getenv(var)
                for var in ("NOTION_SECRET", "NOTION_INTEGRATION_KEY", "NOTION_TOKEN")
                if os.getenv(var)
            ),
            None,
        )
        if token:
            return token
        self.logger.warning("No Notion API token found in environment variables")
        return None

    async def _ensure_initialized(self) -> None:
        if not self._is_initialized or not self.client:
            self.client = httpx.AsyncClient(headers=self.headers, timeout=self.timeout)
            self._is_initialized = True
            self.logger.debug("NotionHttpClient initialized")



from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Self

from notionary.exceptions.search import NoUsersInWorkspace, UserNotFound
from notionary.user.client import UserHttpClient
from notionary.user.schemas import UserResponseDto, UserType
from notionary.utils.fuzzy import find_all_matches

if TYPE_CHECKING:
    from notionary.user.bot import BotUser
    from notionary.user.person import PersonUser


class BaseUser:
    def __init__(
        self,
        id: str,
        name: str | None = None,
        avatar_url: str | None = None,
    ) -> None:
        self._id = id
        self._name = name
        self._avatar_url = avatar_url

    @classmethod
    async def from_id_auto(
        cls,
        user_id: str,
        http_client: UserHttpClient | None = None,
    ) -> BotUser | PersonUser:
        from notionary.user.bot import BotUser
        from notionary.user.person import PersonUser

        client = http_client or UserHttpClient()
        user_dto = await client.get_user_by_id(user_id)

        if user_dto.type == UserType.BOT:
            return BotUser.from_dto(user_dto)
        elif user_dto.type == UserType.PERSON:
            return PersonUser.from_dto(user_dto)
        else:
            raise ValueError(f"Unknown user type: {user_dto.type}")

    @classmethod
    async def from_id(
        cls,
        user_id: str,
        http_client: UserHttpClient | None = None,
    ) -> Self:
        client = http_client or UserHttpClient()
        user_dto = await client.get_user_by_id(user_id)

        expected_type = cls._get_expected_user_type()
        if user_dto.type != expected_type:
            raise ValueError(
                f"User {user_id} is not a '{expected_type.value}', but '{user_dto.type.value}'"
            )

        return cls.from_dto(user_dto)

    @classmethod
    async def from_name(
        cls,
        name: str,
        http_client: UserHttpClient | None = None,
    ) -> Self:
        client = http_client or UserHttpClient()
        all_users = await cls._get_all_users_of_type(client)

        user_type = cls._get_expected_user_type().value

        if not all_users:
            raise NoUsersInWorkspace(user_type)

        exact_match = cls._find_exact_match(all_users, name)
        if exact_match:
            return exact_match

        suggestions = cls._get_fuzzy_suggestions(all_users, name)
        raise UserNotFound(user_type, name, suggestions)

    @classmethod
    def _find_exact_match(cls, users: list[Self], query: str) -> Self | None:
        query_lower = query.lower()
        for user in users:
            if user.name and user.name.lower() == query_lower:
                return user
        return None

    @classmethod
    def _get_fuzzy_suggestions(cls, users: list[Self], query: str) -> list[str]:
        sorted_by_similarity = find_all_matches(
            query=query,
            items=users,
            text_extractor=cls._get_name_extractor(),
            min_similarity=0.6,
        )

        if sorted_by_similarity:
            return [user.name for user in sorted_by_similarity[:5] if user.name]

        return [user.name for user in users[:5] if user.name]

    @classmethod
    async def _get_all_users_of_type(cls, http_client: UserHttpClient) -> list[Self]:
        all_workspace_user_dtos = await http_client.get_all_workspace_users()
        expected_type = cls._get_expected_user_type()
        filtered_dtos = [
            dto for dto in all_workspace_user_dtos if dto.type == expected_type
        ]
        return [cls.from_dto(dto) for dto in filtered_dtos]

    @classmethod
    @abstractmethod
    def _get_expected_user_type(cls) -> UserType:
        pass

    @classmethod
    @abstractmethod
    def from_dto(cls, user_dto: UserResponseDto) -> Self:
        pass

    @classmethod
    def _get_name_extractor(cls):
        return lambda user: user.name or ""

    @property
    def id(self) -> str:
        return self._id

    @property
    def name(self) -> str | None:
        return self._name

    @property
    def avatar_url(self) -> str | None:
        return self._avatar_url

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self._id!r}, name={self._name!r})"



from __future__ import annotations

from enum import StrEnum
from typing import TYPE_CHECKING, Annotated, Literal

from pydantic import BaseModel, ConfigDict, Field

from notionary.blocks.enums import CodingLanguage

if TYPE_CHECKING:
    from notionary.page.content.markdown.structured_output.service import (
        StructuredOutputMarkdownConverter,
    )


class MarkdownNodeType(StrEnum):
    PARAGRAPH = "paragraph"
    HEADING_1 = "heading_1"
    HEADING_2 = "heading_2"
    HEADING_3 = "heading_3"
    SPACE = "space"
    DIVIDER = "divider"
    QUOTE = "quote"
    BULLETED_LIST = "bulleted_list"
    BULLETED_LIST_ITEM = "bulleted_list_item"
    NUMBERED_LIST = "numbered_list"
    NUMBERED_LIST_ITEM = "numbered_list_item"
    TODO = "todo"
    TODO_LIST = "todo_list"
    CALLOUT = "callout"
    TOGGLE = "toggle"
    IMAGE = "image"
    VIDEO = "video"
    AUDIO = "audio"
    FILE = "file"
    PDF = "pdf"
    BOOKMARK = "bookmark"
    EMBED = "embed"
    CODE = "code"
    MERMAID = "mermaid"
    TABLE = "table"
    BREADCRUMB = "breadcrumb"
    EQUATION = "equation"
    TABLE_OF_CONTENTS = "table_of_contents"
    COLUMNS = "columns"


class MarkdownNodeSchema(BaseModel):
    type: MarkdownNodeType

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        raise NotImplementedError(
            f"{self.__class__.__name__} must implement process_with()"
        )


class ParagraphSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.PARAGRAPH] = MarkdownNodeType.PARAGRAPH
    text: str = Field(description="The paragraph text content")

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_paragraph(self)


class Heading1Schema(MarkdownNodeSchema):
    type: Literal["heading_1"] = "heading_1"
    text: str = Field(description="The heading 1 text")
    children: list[MarkdownNodeSchema] | None = Field(
        default=None, description="Optional child nodes"
    )

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_heading_1(self)


class Heading2Schema(MarkdownNodeSchema):
    type: Literal["heading_2"] = "heading_2"
    text: str = Field(description="The heading 2 text")
    children: list[MarkdownNodeSchema] | None = Field(
        default=None, description="Optional child nodes"
    )

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_heading_2(self)


class Heading3Schema(MarkdownNodeSchema):
    type: Literal["heading_3"] = "heading_3"
    text: str = Field(description="The heading 3 text")
    children: list[MarkdownNodeSchema] | None = Field(
        default=None, description="Optional child nodes"
    )

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_heading_3(self)


class SpaceSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.SPACE] = MarkdownNodeType.SPACE

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_space()


class DividerSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.DIVIDER] = MarkdownNodeType.DIVIDER

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_divider()


class QuoteSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.QUOTE] = MarkdownNodeType.QUOTE
    text: str = Field(description="The quote text")
    children: list[MarkdownNodeSchema] | None = Field(
        default=None, description="Optional child nodes"
    )

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_quote(self)


class BulletedListItemSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.BULLETED_LIST_ITEM] = (
        MarkdownNodeType.BULLETED_LIST_ITEM
    )
    text: str = Field(description="The bullet point text")
    children: list[MarkdownNodeSchema] | None = Field(
        default=None, description="Optional nested content"
    )

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_bulleted_list_item(self)


class BulletedListSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.BULLETED_LIST] = MarkdownNodeType.BULLETED_LIST
    items: list[BulletedListItemSchema] = Field(
        description="List of BulletedListItemSchema objects. Each item must have 'type', 'text', and optionally 'children'"
    )

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_bulleted_list(self)


class NumberedListItemSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.NUMBERED_LIST_ITEM] = (
        MarkdownNodeType.NUMBERED_LIST_ITEM
    )
    text: str = Field(description="The numbered item text")
    children: list[MarkdownNodeSchema] | None = Field(
        default=None, description="Optional nested content"
    )

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_numbered_list_item(self)


class NumberedListSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.NUMBERED_LIST] = MarkdownNodeType.NUMBERED_LIST
    items: list[NumberedListItemSchema] = Field(
        description="List of NumberedListItemSchema objects. Each item must have 'type', 'text', and optionally 'children'"
    )

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_numbered_list(self)


class TodoSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.TODO] = MarkdownNodeType.TODO
    text: str = Field(description="The todo item text")
    checked: bool = Field(default=False, description="Whether the todo is completed")
    children: list[MarkdownNodeSchema] | None = Field(
        default=None, description="Optional nested content"
    )

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_todo(self)


class TodoListSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.TODO_LIST] = MarkdownNodeType.TODO_LIST
    items: list[TodoSchema] = Field(
        description="List of TodoSchema objects. Each item must have 'type', 'text', 'checked', and optionally 'children'"
    )

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_todo_list(self)


class CalloutSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.CALLOUT] = MarkdownNodeType.CALLOUT
    text: str = Field(description="The callout text")
    emoji: str | None = Field(default=None, description="Optional emoji icon")
    children: list[MarkdownNodeSchema] | None = Field(
        default=None, description="Optional child nodes"
    )

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_callout(self)


class ToggleSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.TOGGLE] = MarkdownNodeType.TOGGLE
    title: str = Field(description="The toggle title")
    children: list[MarkdownNodeSchema] = Field(description="Content inside the toggle")

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_toggle(self)


class ImageSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.IMAGE] = MarkdownNodeType.IMAGE
    url: str = Field(description="Image URL")
    caption: str | None = Field(default=None, description="Optional caption")

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_image(self)


class VideoSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.VIDEO] = MarkdownNodeType.VIDEO
    url: str = Field(description="Video URL")
    caption: str | None = Field(default=None, description="Optional caption")

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_video(self)


class AudioSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.AUDIO] = MarkdownNodeType.AUDIO
    url: str = Field(description="Audio URL")
    caption: str | None = Field(default=None, description="Optional caption")

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_audio(self)


class FileSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.FILE] = MarkdownNodeType.FILE
    url: str = Field(description="File URL")
    caption: str | None = Field(default=None, description="Optional caption")

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_file(self)


class PdfSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.PDF] = MarkdownNodeType.PDF
    url: str = Field(description="PDF URL")
    caption: str | None = Field(default=None, description="Optional caption")

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_pdf(self)


class BookmarkSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.BOOKMARK] = MarkdownNodeType.BOOKMARK
    url: str = Field(description="Bookmark URL")
    title: str | None = Field(default=None, description="Optional title")
    caption: str | None = Field(default=None, description="Optional caption")

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_bookmark(self)


class EmbedSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.EMBED] = MarkdownNodeType.EMBED
    url: str = Field(description="Embed URL")
    caption: str | None = Field(default=None, description="Optional caption")

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_embed(self)


class CodeSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.CODE] = MarkdownNodeType.CODE
    code: str = Field(description="Code content")
    language: CodingLanguage | None = Field(
        default=None, description="Programming language"
    )
    caption: str | None = Field(default=None, description="Optional caption")

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_code(self)


class MermaidSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.MERMAID] = MarkdownNodeType.MERMAID
    diagram: str = Field(description="Mermaid diagram code")
    caption: str | None = Field(default=None, description="Optional caption")

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_mermaid(self)


class TableSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.TABLE] = MarkdownNodeType.TABLE
    headers: list[str] = Field(description="Table header row")
    rows: list[list[str]] = Field(description="Table data rows")

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_table(self)


class BreadcrumbSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.BREADCRUMB] = MarkdownNodeType.BREADCRUMB

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_breadcrumb(self)


class EquationSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.EQUATION] = MarkdownNodeType.EQUATION
    expression: str = Field(description="LaTeX equation expression")

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_equation(self)


class TableOfContentsSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.TABLE_OF_CONTENTS] = (
        MarkdownNodeType.TABLE_OF_CONTENTS
    )

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_table_of_contents(self)


class ColumnSchema(BaseModel):
    """Single column in a multi-column layout."""

    width_ratio: float | None = Field(
        default=None,
        description="Relative width of this column (e.g., 0.5 for half width). If not specified, columns are equal width",
    )
    children: list[MarkdownNodeSchema] = Field(
        description="Content inside this column. Can contain any markdown nodes"
    )


class ColumnsSchema(MarkdownNodeSchema):
    type: Literal[MarkdownNodeType.COLUMNS] = MarkdownNodeType.COLUMNS
    columns: list[ColumnSchema] = Field(
        description="List of columns in this layout. Each column contains its own content"
    )

    def process_with(self, processor: StructuredOutputMarkdownConverter) -> None:
        processor._process_columns(self)


type AnyMarkdownNode = Annotated[
    Heading1Schema
    | Heading2Schema
    | Heading3Schema
    | ParagraphSchema
    | SpaceSchema
    | DividerSchema
    | QuoteSchema
    | BulletedListSchema
    | BulletedListItemSchema
    | NumberedListSchema
    | NumberedListItemSchema
    | TodoSchema
    | TodoListSchema
    | CalloutSchema
    | ToggleSchema
    | ImageSchema
    | VideoSchema
    | AudioSchema
    | FileSchema
    | PdfSchema
    | BookmarkSchema
    | EmbedSchema
    | CodeSchema
    | MermaidSchema
    | TableSchema
    | BreadcrumbSchema
    | EquationSchema
    | TableOfContentsSchema
    | ColumnsSchema,
    Field(discriminator="type"),
]


class MarkdownDocumentSchema(BaseModel):
    model_config = ConfigDict(extra="forbid")

    nodes: list[AnyMarkdownNode] = Field(
        description="Ordered list of top-level markdown nodes in the document. Each node can contain nested children"
    )



from notionary.blocks.client import NotionBlockHttpClient
from notionary.page.content.parser.factory import ConverterChainFactory
from notionary.page.content.parser.post_processing.handlers import (
    RichTextLengthTruncationPostProcessor,
)
from notionary.page.content.parser.post_processing.service import BlockPostProcessor
from notionary.page.content.parser.pre_processsing.handlers import (
    ColumnSyntaxPreProcessor,
    IndentationNormalizer,
    VideoFormatPreProcessor,
    WhitespacePreProcessor,
)
from notionary.page.content.parser.pre_processsing.service import MarkdownPreProcessor
from notionary.page.content.parser.service import MarkdownToNotionConverter
from notionary.page.content.renderer.factory import RendererChainFactory
from notionary.page.content.renderer.post_processing.handlers import (
    NumberedListPlaceholderReplacerPostProcessor,
)
from notionary.page.content.renderer.post_processing.service import (
    MarkdownRenderingPostProcessor,
)
from notionary.page.content.renderer.service import NotionToMarkdownConverter
from notionary.page.content.service import PageContentService


class PageContentServiceFactory:
    def __init__(
        self,
        converter_chain_factory: ConverterChainFactory | None = None,
        renderer_chain_factory: RendererChainFactory | None = None,
    ) -> None:
        self._converter_chain_factory = (
            converter_chain_factory or ConverterChainFactory()
        )
        self._renderer_chain_factory = renderer_chain_factory or RendererChainFactory()

    def create(
        self, page_id: str, block_client: NotionBlockHttpClient
    ) -> PageContentService:
        markdown_converter = self._create_markdown_to_notion_converter()
        notion_to_markdown_converter = self._create_notion_to_markdown_converter()

        return PageContentService(
            page_id=page_id,
            block_client=block_client,
            markdown_converter=markdown_converter,
            notion_to_markdown_converter=notion_to_markdown_converter,
        )

    def _create_markdown_to_notion_converter(self) -> MarkdownToNotionConverter:
        line_parser = self._converter_chain_factory.create()
        markdown_pre_processor = self._create_markdown_preprocessor()
        block_post_processor = self._create_post_processor()

        return MarkdownToNotionConverter(
            line_parser=line_parser,
            pre_processor=markdown_pre_processor,
            post_processor=block_post_processor,
        )

    def _create_notion_to_markdown_converter(self) -> NotionToMarkdownConverter:
        renderer_chain = self._renderer_chain_factory.create()
        markdown_rendering_post_processor = (
            self._create_markdown_rendering_post_processor()
        )
        return NotionToMarkdownConverter(
            renderer_chain=renderer_chain,
            post_processor=markdown_rendering_post_processor,
        )

    def _create_markdown_preprocessor(self) -> MarkdownPreProcessor:
        pre_processor = MarkdownPreProcessor()
        pre_processor.register(ColumnSyntaxPreProcessor())
        pre_processor.register(WhitespacePreProcessor())
        pre_processor.register(IndentationNormalizer())
        pre_processor.register(VideoFormatPreProcessor())
        return pre_processor

    def _create_post_processor(self) -> BlockPostProcessor:
        post_processor = BlockPostProcessor()
        post_processor.register(RichTextLengthTruncationPostProcessor())
        return post_processor

    def _create_markdown_rendering_post_processor(
        self,
    ) -> MarkdownRenderingPostProcessor:
        post_processor = MarkdownRenderingPostProcessor()
        post_processor.register(NumberedListPlaceholderReplacerPostProcessor())
        return post_processor



from typing import override

from notionary.blocks.enums import BlockType
from notionary.blocks.rich_text.rich_text_markdown_converter import (
    RichTextToMarkdownConverter,
)
from notionary.blocks.schemas import Block
from notionary.page.content.renderer.context import MarkdownRenderingContext
from notionary.page.content.renderer.renderers.base import BlockRenderer
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class TodoRenderer(BlockRenderer):
    def __init__(
        self,
        syntax_registry: SyntaxDefinitionRegistry | None = None,
        rich_text_markdown_converter: RichTextToMarkdownConverter | None = None,
    ) -> None:
        super().__init__(syntax_registry=syntax_registry)
        self._rich_text_markdown_converter = (
            rich_text_markdown_converter or RichTextToMarkdownConverter()
        )

    @override
    def _can_handle(self, block: Block) -> bool:
        return block.type == BlockType.TO_DO

    @override
    async def _process(self, context: MarkdownRenderingContext) -> None:
        is_checked, content = await self._extract_todo_info(context.block)

        if not content:
            context.markdown_result = ""
            return

        syntax = (
            self._syntax_registry.get_todo_done_syntax()
            if is_checked
            else self._syntax_registry.get_todo_syntax()
        )

        todo_markdown = f"{syntax.start_delimiter} {content}"

        if context.indent_level > 0:
            todo_markdown = context.indent_text(todo_markdown)

        children_markdown = await context.render_children_with_additional_indent(1)

        if children_markdown:
            context.markdown_result = f"{todo_markdown}\n{children_markdown}"
        else:
            context.markdown_result = todo_markdown

    async def _extract_todo_info(self, block: Block) -> tuple[bool, str]:
        if not block.to_do:
            return False, ""

        is_checked = block.to_do.checked or False

        content = ""
        if block.to_do.rich_text:
            content = await self._rich_text_markdown_converter.to_markdown(
                block.to_do.rich_text
            )

        return is_checked, content



from typing import override

from notionary.page.content.markdown.nodes.base import MarkdownNode
from notionary.page.content.markdown.nodes.container import ContainerNode
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class TodoMarkdownNode(ContainerNode):
    VALID_MARKER = "-"

    def __init__(
        self,
        text: str,
        checked: bool = False,
        marker: str = "-",
        children: list[MarkdownNode] | None = None,
        syntax_registry: SyntaxDefinitionRegistry | None = None,
    ):
        super().__init__(syntax_registry=syntax_registry)
        self.text = text
        self.checked = checked
        self.marker = marker
        self.children = children or []

    @override
    def to_markdown(self) -> str:
        # Get the appropriate syntax based on checked state
        if self.checked:
            todo_syntax = self._syntax_registry.get_todo_done_syntax()
        else:
            todo_syntax = self._syntax_registry.get_todo_syntax()

        result = f"{todo_syntax.start_delimiter} {self.text}"
        result += self.render_children()
        return result

    def _get_validated_marker(self) -> str:
        return self.marker if self.marker == self.VALID_MARKER else self.VALID_MARKER

    def _get_checkbox_state(self) -> str:
        if self.checked:
            todo_done_syntax = self._syntax_registry.get_todo_done_syntax()
            return todo_done_syntax.start_delimiter
        else:
            todo_syntax = self._syntax_registry.get_todo_syntax()
            return todo_syntax.start_delimiter



from typing import override

from notionary.page.content.markdown.nodes.base import MarkdownNode
from notionary.page.content.markdown.nodes.container import ContainerNode
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class ColumnMarkdownNode(ContainerNode):
    def __init__(
        self,
        children: list[MarkdownNode] | None = None,
        width_ratio: float | None = None,
        syntax_registry: SyntaxDefinitionRegistry | None = None,
    ):
        super().__init__(syntax_registry=syntax_registry)
        self.children = children or []
        self.width_ratio = width_ratio

    @override
    def to_markdown(self) -> str:
        start_tag = self._format_column_start_tag()
        result = start_tag + self.render_children()
        return result

    def _format_column_start_tag(self) -> str:
        delimiter = self._syntax_registry.get_column_syntax().start_delimiter

        if self.width_ratio is not None:
            return f"{delimiter} {self.width_ratio}"
        return delimiter


class ColumnListMarkdownNode(MarkdownNode):
    def __init__(
        self,
        columns: list[ColumnMarkdownNode] | None = None,
        syntax_registry: SyntaxDefinitionRegistry | None = None,
    ):
        super().__init__(syntax_registry=syntax_registry)
        self.columns = columns or []

    @override
    def to_markdown(self) -> str:
        start_delimiter = self._get_column_list_delimiter()

        if not self.columns:
            return start_delimiter

        result = start_delimiter + self._render_columns()
        return result

    def _get_column_list_delimiter(self) -> str:
        return self._syntax_registry.get_column_list_syntax().start_delimiter

    def _render_columns(self) -> str:
        rendered_columns = []

        for column in self.columns:
            column_markdown = column.to_markdown()
            if column_markdown:
                indented = self._indent_column(column_markdown)
                rendered_columns.append(indented)

        return "\n" + "\n".join(rendered_columns) if rendered_columns else ""

    @staticmethod
    def _indent_column(text: str, indent: str = "    ") -> str:
        lines = text.split("\n")
        return "\n".join(f"{indent}{line}" if line.strip() else line for line in lines)



from pathlib import Path
from typing import override

from notionary.exceptions.file_upload import FileNotFoundError
from notionary.file_upload.validation.port import FileUploadValidator


class FileExistsValidator(FileUploadValidator):
    def __init__(self, file_path: Path) -> None:
        self.file_path = file_path

    @override
    async def validate(self) -> None:
        if not self.file_path.exists():
            raise FileNotFoundError(str(self.file_path))



from typing import override

from notionary.blocks.rich_text.rich_text_markdown_converter import (
    RichTextToMarkdownConverter,
)
from notionary.blocks.schemas import Block, BlockType, HeadingData
from notionary.page.content.renderer.context import MarkdownRenderingContext
from notionary.page.content.renderer.renderers.base import BlockRenderer
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class HeadingRenderer(BlockRenderer):
    MIN_HEADING_LEVEL = 1
    MAX_HEADING_LEVEL = 3

    def __init__(
        self,
        syntax_registry: SyntaxDefinitionRegistry | None = None,
        rich_text_markdown_converter: RichTextToMarkdownConverter | None = None,
    ) -> None:
        super().__init__(syntax_registry=syntax_registry)
        self._syntax = self._syntax_registry.get_heading_syntax()
        self._rich_text_markdown_converter = (
            rich_text_markdown_converter or RichTextToMarkdownConverter()
        )

    @override
    def _can_handle(self, block: Block) -> bool:
        return block.type in (
            BlockType.HEADING_1,
            BlockType.HEADING_2,
            BlockType.HEADING_3,
        )

    @override
    async def _process(self, context: MarkdownRenderingContext) -> None:
        level = self._get_heading_level(context.block)
        title = await self._get_heading_title(context.block)

        if not self._is_valid_heading(level, title):
            return

        heading_markdown = self._format_heading(level, title, context.indent_level)

        if self._is_toggleable(context.block):
            context.markdown_result = await self._render_toggleable_heading(
                heading_markdown, context
            )
        else:
            context.markdown_result = heading_markdown

    def _is_valid_heading(self, level: int, title: str) -> bool:
        return self.MIN_HEADING_LEVEL <= level <= self.MAX_HEADING_LEVEL and bool(title)

    def _format_heading(self, level: int, title: str, indent_level: int) -> str:
        heading_prefix = self._syntax.start_delimiter * level
        heading_markdown = f"{heading_prefix} {title}"

        if indent_level > 0:
            indent = "    " * indent_level
            heading_markdown = f"{indent}{heading_markdown}"

        return heading_markdown

    async def _render_toggleable_heading(
        self, heading_markdown: str, context: MarkdownRenderingContext
    ) -> str:
        original_indent = context.indent_level
        context.indent_level += 1

        children_markdown = await context.render_children()

        context.indent_level = original_indent

        if children_markdown:
            return f"{heading_markdown}\n{children_markdown}"
        return heading_markdown

    def _get_heading_level(self, block: Block) -> int:
        if block.type == BlockType.HEADING_1:
            return 1
        elif block.type == BlockType.HEADING_2:
            return 2
        elif block.type == BlockType.HEADING_3:
            return 3
        return 0

    def _is_toggleable(self, block: Block) -> bool:
        heading_data = self._get_heading_data(block)
        return heading_data.is_toggleable if heading_data else False

    async def _get_heading_title(self, block: Block) -> str:
        heading_data = self._get_heading_data(block)

        if not heading_data or not heading_data.rich_text:
            return ""

        return await self._rich_text_markdown_converter.to_markdown(
            heading_data.rich_text
        )

    def _get_heading_data(self, block: Block) -> HeadingData | None:
        if block.type == BlockType.HEADING_1:
            return block.heading_1
        elif block.type == BlockType.HEADING_2:
            return block.heading_2
        elif block.type == BlockType.HEADING_3:
            return block.heading_3
        return None



from collections.abc import AsyncGenerator

from notionary.blocks.rich_text.models import RichText
from notionary.comments.schemas import (
    CommentCreateRequest,
    CommentDto,
    CommentListRequest,
    CommentListResponse,
)
from notionary.http.client import NotionHttpClient
from notionary.utils.pagination import (
    paginate_notion_api,
    paginate_notion_api_generator,
)


class CommentClient(NotionHttpClient):
    def __init__(self) -> None:
        super().__init__()

    async def iter_comments(
        self,
        block_id: str,
        total_results_limit: int | None = None,
    ) -> AsyncGenerator[CommentDto]:
        async for comment in paginate_notion_api_generator(
            self._list_comments_page,
            block_id=block_id,
            total_results_limit=total_results_limit,
        ):
            yield comment

    async def get_all_comments(
        self, block_id: str, *, total_results_limit: int | None = None
    ) -> list[CommentDto]:
        all_comments = await paginate_notion_api(
            self._list_comments_page,
            block_id=block_id,
            total_results_limit=total_results_limit,
        )

        self.logger.debug(
            "Retrieved %d total comments for block %s", len(all_comments), block_id
        )
        return all_comments

    async def _list_comments_page(
        self,
        block_id: str,
        *,
        start_cursor: str | None = None,
        page_size: int = 100,
    ) -> CommentListResponse:
        request = CommentListRequest(
            block_id=block_id,
            start_cursor=start_cursor,
            page_size=page_size,
        )
        resp = await self.get("comments", params=request.model_dump(exclude_none=True))
        return CommentListResponse.model_validate(resp)

    async def create_comment_for_page(
        self,
        rich_text: list[RichText],
        page_id: str,
    ) -> CommentDto:
        request = CommentCreateRequest.for_page(page_id=page_id, rich_text=rich_text)

        body = request.model_dump(exclude_unset=True, exclude_none=True)

        resp = await self.post("comments", data=body)
        return CommentDto.model_validate(resp)

    async def create_comment_for_discussion(
        self,
        rich_text: list[RichText],
        discussion_id: str,
    ) -> CommentDto:
        request = CommentCreateRequest.for_discussion(
            discussion_id=discussion_id, rich_text=rich_text
        )

        body = request.model_dump(exclude_unset=True, exclude_none=True)

        resp = await self.post("comments", data=body)
        return CommentDto.model_validate(resp)



from typing import Generic, Literal, TypeVar

from pydantic import BaseModel

from notionary.data_source.schemas import DataSourceDto
from notionary.page.schemas import NotionPageDto

T = TypeVar("T", bound=BaseModel)

PageOrDataSource = DataSourceDto | NotionPageDto


class SearchResponse(BaseModel, Generic[T]):
    results: list[T]
    next_cursor: str | None = None
    has_more: bool
    type: Literal["page_or_data_source"]


PageSearchResponse = SearchResponse[NotionPageDto]
DataSourceSearchResponse = SearchResponse[DataSourceDto]



from abc import abstractmethod
from typing import override

from notionary.blocks.rich_text.rich_text_markdown_converter import (
    RichTextToMarkdownConverter,
)
from notionary.blocks.schemas import Block
from notionary.page.content.renderer.context import MarkdownRenderingContext
from notionary.page.content.renderer.renderers.base import BlockRenderer
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class CaptionedBlockRenderer(BlockRenderer):
    def __init__(
        self,
        syntax_registry: SyntaxDefinitionRegistry | None = None,
        rich_text_markdown_converter: RichTextToMarkdownConverter | None = None,
    ) -> None:
        super().__init__(syntax_registry=syntax_registry)
        self._rich_text_markdown_converter = (
            rich_text_markdown_converter or RichTextToMarkdownConverter()
        )

    @abstractmethod
    async def _render_main_content(self, block: Block) -> str:
        raise NotImplementedError

    @override
    async def _process(self, context: MarkdownRenderingContext) -> None:
        main_content = await self._render_main_content(context.block)

        if not main_content:
            context.markdown_result = ""
            return

        caption_markdown = await self._render_caption(context.block)

        final_markdown = f"{main_content}{caption_markdown}"

        if context.indent_level > 0:
            final_markdown = context.indent_text(final_markdown)

        children_markdown = await context.render_children_with_additional_indent(1)

        if children_markdown:
            context.markdown_result = f"{final_markdown}\n{children_markdown}"
        else:
            context.markdown_result = final_markdown

    async def _render_caption(self, block: Block) -> str:
        block_data_object = getattr(block, block.type.value, None)

        if not block_data_object or not hasattr(block_data_object, "caption"):
            return ""

        caption_rich_text = getattr(block_data_object, "caption", [])
        if not caption_rich_text:
            return ""

        caption_markdown = await self._rich_text_markdown_converter.to_markdown(
            caption_rich_text
        )

        return f"\n[caption] {caption_markdown}"



from typing import override

from notionary.blocks.rich_text.name_id_resolver.port import NameIdResolver
from notionary.workspace.query.service import WorkspaceQueryService


# !!! in the notion api mentions that reference datasources are not provided yet (it's a limiation of the API as of now)
class DataSourceNameIdResolver(NameIdResolver):
    def __init__(
        self, workspace_query_service: WorkspaceQueryService | None = None
    ) -> None:
        self._workspace_query_service = (
            workspace_query_service or WorkspaceQueryService()
        )

    @override
    async def resolve_name_to_id(self, name: str) -> str | None:
        if not name:
            return None

        cleaned_name = name.strip()
        data_source = await self._workspace_query_service.find_data_source(
            query=cleaned_name
        )
        return data_source.id if data_source else None

    @override
    async def resolve_id_to_name(self, data_source_id: str) -> str | None:
        if not data_source_id:
            return None

        try:
            from notionary import NotionDataSource

            data_source = await NotionDataSource.from_id(data_source_id)
            return data_source.title if data_source else None
        except Exception:
            return None



from collections.abc import AsyncGenerator
from pathlib import Path

import aiofiles

from notionary.file_upload.config.config import FileUploadConfig


class FileContentReader:
    def __init__(self, config: FileUploadConfig | None = None):
        config = config or FileUploadConfig()
        self._chunk_size = config.multi_part_chunk_size

    async def read_full_file(self, file_path: Path) -> bytes:
        async with aiofiles.open(file_path, "rb") as f:
            return await f.read()

    async def read_file_chunks(self, file_path: Path) -> AsyncGenerator[bytes]:
        async with aiofiles.open(file_path, "rb") as file:
            while True:
                chunk = await file.read(self._chunk_size)
                if not chunk:
                    break
                yield chunk

    async def bytes_to_chunks(self, file_content: bytes) -> AsyncGenerator[bytes]:
        for i in range(0, len(file_content), self._chunk_size):
            yield file_content[i : i + self._chunk_size]



from pydantic import BaseModel

from notionary.blocks.rich_text.models import RichText
from notionary.data_source.properties.schemas import AnyDataSourceProperty
from notionary.page.schemas import NotionPageDto
from notionary.shared.entity.schemas import EntityResponseDto, NotionEntityUpdateDto
from notionary.shared.models.parent import Parent


class UpdateDataSourceDto(NotionEntityUpdateDto):
    title: list[RichText]
    description: list[RichText]
    archived: bool


class QueryDataSourceResponse(BaseModel):
    results: list[NotionPageDto]
    next_cursor: str | None = None
    has_more: bool


class DataSourceDto(EntityResponseDto):
    database_parent: Parent
    title: list[RichText]
    description: list[RichText]
    archived: bool
    properties: dict[str, AnyDataSourceProperty]



from typing import cast

from notionary.page.properties.client import PagePropertyHttpClient
from notionary.page.properties.service import PagePropertyHandler
from notionary.page.schemas import NotionPageDto
from notionary.shared.models.parent import DataSourceParent, ParentType


class PagePropertyHandlerFactory:
    def create_from_page_response(
        self, page_response: NotionPageDto
    ) -> PagePropertyHandler:
        return PagePropertyHandler(
            properties=page_response.properties,
            parent_type=page_response.parent.type,
            page_url=page_response.url,
            page_property_http_client=self._create_http_client(
                page_id=page_response.id
            ),
            parent_data_source=self._extract_parent_data_source_id(page_response),
        )

    def _create_http_client(self, page_id: str) -> PagePropertyHttpClient:
        return PagePropertyHttpClient(page_id=page_id)

    def _extract_parent_data_source_id(self, response: NotionPageDto) -> str | None:
        if response.parent.type != ParentType.DATA_SOURCE_ID:
            return None
        data_source_parent = cast(DataSourceParent, response.parent)
        return data_source_parent.data_source_id if data_source_parent else None



from typing import Any

from notionary.data_source.query.schema import (
    ArrayOperator,
    BooleanOperator,
    DateOperator,
    NumberOperator,
    Operator,
    StringOperator,
)
from notionary.exceptions.base import NotionaryException
from notionary.shared.properties.type import PropertyType


class QueryBuilderError(NotionaryException):
    def __init__(self, message: str, property_name: str | None = None) -> None:
        self.property_name = property_name
        super().__init__(message)


class InvalidOperatorForPropertyType(QueryBuilderError):
    def __init__(
        self,
        property_name: str,
        property_type: PropertyType,
        operator: Operator,
        valid_operators: list[Operator],
    ) -> None:
        self.property_type = property_type
        self.operator = operator
        self.valid_operators = valid_operators

        valid_operators_str = ", ".join([f"'{op.value}'" for op in valid_operators])
        message = (
            f"Cannot use operator '{operator.value}' on property '{property_name}' of type '{property_type}'. "
            f"Valid operators for this property type are: {valid_operators_str}"
        )
        super().__init__(message, property_name)


class InvalidValueForOperator(QueryBuilderError):
    def __init__(
        self,
        property_name: str,
        operator: Operator,
        provided_value: Any,
        expected_value_type: str,
        example_value: Any | None = None,
    ) -> None:
        self.operator = operator
        self.provided_value = provided_value
        self.expected_value_type = expected_value_type
        self.example_value = example_value

        message = (
            f"Invalid value for operator '{operator.value}' on property '{property_name}'. "
            f"Expected {expected_value_type}, but got {type(provided_value).__name__}"
        )

        if example_value is not None:
            message += f". Example: {example_value}"

        super().__init__(message, property_name)


class MissingRequiredValue(QueryBuilderError):
    def __init__(
        self,
        property_name: str,
        operator: StringOperator
        | NumberOperator
        | BooleanOperator
        | DateOperator
        | ArrayOperator,
    ) -> None:
        self.operator = operator

        message = f"Operator '{operator.value}' on property '{property_name}' requires a value, but none was provided"
        super().__init__(message, property_name)


class ValueNotAllowedForOperator(QueryBuilderError):
    def __init__(
        self,
        property_name: str,
        operator: StringOperator
        | NumberOperator
        | BooleanOperator
        | DateOperator
        | ArrayOperator,
    ) -> None:
        self.operator = operator

        message = (
            f"Operator '{operator.value}' on property '{property_name}' does not accept a value. "
            f"Operators like 'is_empty', 'is_not_empty', 'is_true', 'is_false' don't need values"
        )
        super().__init__(message, property_name)


class InvalidDateFormat(QueryBuilderError):
    def __init__(
        self,
        property_name: str,
        provided_date: str,
        expected_format: str = "ISO 8601 (YYYY-MM-DD or YYYY-MM-DDTHH:MM:SS)",
    ) -> None:
        self.provided_date = provided_date
        self.expected_format = expected_format

        message = (
            f"Invalid date format for property '{property_name}'. "
            f"Provided: '{provided_date}', expected format: {expected_format}"
        )
        super().__init__(message, property_name)


class InvalidNumberValue(QueryBuilderError):
    def __init__(
        self,
        property_name: str,
        provided_value: Any,
        reason: str,
    ) -> None:
        self.provided_value = provided_value
        self.reason = reason

        message = f"Invalid number value for property '{property_name}': {reason}. Provided value: {provided_value}"
        super().__init__(message, property_name)


class PropertyNotConfigured(QueryBuilderError):
    def __init__(
        self,
        property_name: str,
    ) -> None:
        message = (
            f"Property '{property_name}' exists but has no type configuration. "
            f"Cannot determine valid operators for this property"
        )
        super().__init__(message, property_name)


class NoPropertySelected(QueryBuilderError):
    def __init__(self) -> None:
        message = (
            "No property selected. Use .where('property_name') or .where_not('property_name') "
            "before calling an operator method like .equals(), .contains(), etc."
        )
        super().__init__(message)


class EmptyOrGroupError(QueryBuilderError):
    def __init__(self) -> None:
        message = "Cannot create an OR group with no conditions. Add at least one filter condition before using .or_where()"
        super().__init__(message)


class ConflictingFiltersError(QueryBuilderError):
    def __init__(
        self,
        property_name: str,
        description: str,
    ) -> None:
        message = (
            f"Conflicting filters detected on property '{property_name}': {description}. "
            f"This query will never return any results"
        )
        super().__init__(message, property_name)


class InvalidSortProperty(QueryBuilderError):
    def __init__(
        self,
        property_name: str,
        reason: str,
        available_properties: list[str] | None = None,
    ) -> None:
        self.available_properties = available_properties
        self.reason = reason

        message = f"Cannot sort by property '{property_name}': {reason}"

        if available_properties:
            props_str = ", ".join([f"'{p}'" for p in sorted(available_properties)])
            message += f". Available properties: {props_str}"

        super().__init__(message, property_name)



from pathlib import Path

from notionary.file_upload.validation.service import FileUploadValidationService
from notionary.file_upload.validation.validators import (
    FileExistsValidator,
    FileExtensionValidator,
    FileNameLengthValidator,
    FileUploadLimitValidator,
)


def create_file_upload_validation_service(
    file_path: Path,
) -> FileUploadValidationService:
    file_path = Path(file_path)
    filename = file_path.name
    file_size_bytes = file_path.stat().st_size

    validation_service = FileUploadValidationService()

    file_exists_validator = _create_file_exists_validator(file_path)
    filename_length_validator = _create_filename_length_validator(filename)
    extension_validator = _create_extension_validator(filename)
    size_validator = _create_size_validator(filename, file_size_bytes)

    validation_service.register(file_exists_validator)
    validation_service.register(filename_length_validator)
    validation_service.register(extension_validator)
    validation_service.register(size_validator)

    return validation_service


def create_bytes_upload_validation_service(
    filename: str,
    file_size_bytes: int,
) -> FileUploadValidationService:
    validation_service = FileUploadValidationService()

    filename_length_validator = _create_filename_length_validator(filename)
    extension_validator = _create_extension_validator(filename)
    size_validator = _create_size_validator(filename, file_size_bytes)

    validation_service.register(filename_length_validator)
    validation_service.register(extension_validator)
    validation_service.register(size_validator)

    return validation_service


def _create_file_exists_validator(file_path: Path) -> FileExistsValidator:
    return FileExistsValidator(file_path=file_path)


def _create_filename_length_validator(filename: str) -> FileNameLengthValidator:
    return FileNameLengthValidator(filename=filename)


def _create_extension_validator(filename: str) -> FileExtensionValidator:
    return FileExtensionValidator(filename=filename)


def _create_size_validator(
    filename: str, file_size_bytes: int
) -> FileUploadLimitValidator:
    return FileUploadLimitValidator(filename=filename, file_size_bytes=file_size_bytes)



from enum import StrEnum
from typing import Protocol

from pydantic import BaseModel

from notionary.blocks.rich_text.models import RichText
from notionary.shared.models.file import File
from notionary.shared.models.icon import Icon
from notionary.shared.models.parent import Parent
from notionary.user.schemas import PartialUserDto


class _EntityType(StrEnum):
    PAGE = "page"
    DATA_SOURCE = "data_source"
    DATABASE = "database"


class EntityResponseDto(BaseModel):
    object: _EntityType
    id: str
    created_time: str
    created_by: PartialUserDto
    last_edited_time: str
    last_edited_by: PartialUserDto
    cover: File | None = None
    icon: Icon | None = None
    parent: Parent
    in_trash: bool
    url: str
    public_url: str | None = None


class NotionEntityUpdateDto(BaseModel):
    icon: Icon | None = None
    cover: File | None = None
    in_trash: bool | None = None


class Titled(Protocol):
    title: list[RichText]


class Describable(Protocol):
    description: list[RichText] | None



from collections.abc import AsyncGenerator

from notionary.data_source.schemas import DataSourceDto
from notionary.http.client import NotionHttpClient
from notionary.page.schemas import NotionPageDto
from notionary.shared.typings import JsonDict
from notionary.utils.pagination import paginate_notion_api_generator
from notionary.workspace.query.models import WorkspaceQueryConfig
from notionary.workspace.schemas import DataSourceSearchResponse, PageSearchResponse


class WorkspaceClient:
    DEFAULT_PAGE_SIZE = 100

    def __init__(self, http_client: NotionHttpClient | None = None) -> None:
        self._http_client = http_client or NotionHttpClient()

    async def query_pages_stream(
        self,
        search_config: WorkspaceQueryConfig,
    ) -> AsyncGenerator[NotionPageDto]:
        async for page in paginate_notion_api_generator(
            self._query_pages,
            search_config=search_config,
            total_results_limit=search_config.total_results_limit,
        ):
            yield page

    async def query_data_sources_stream(
        self,
        search_config: WorkspaceQueryConfig,
    ) -> AsyncGenerator[DataSourceDto]:
        async for data_source in paginate_notion_api_generator(
            self._query_data_sources,
            search_config=search_config,
            total_results_limit=search_config.total_results_limit,
        ):
            yield data_source

    async def _query_pages(
        self,
        search_config: WorkspaceQueryConfig,
        start_cursor: str | None = None,
    ) -> PageSearchResponse:
        if start_cursor:
            search_config.start_cursor = start_cursor

        response = await self._execute_search(search_config)
        return PageSearchResponse.model_validate(response)

    async def _query_data_sources(
        self,
        search_config: WorkspaceQueryConfig,
        start_cursor: str | None = None,
    ) -> DataSourceSearchResponse:
        if start_cursor:
            search_config.start_cursor = start_cursor

        response = await self._execute_search(search_config)
        return DataSourceSearchResponse.model_validate(response)

    async def _execute_search(self, config: WorkspaceQueryConfig) -> JsonDict:
        serialized_config = config.model_dump(exclude_none=True, by_alias=True)
        return await self._http_client.post("search", serialized_config)



from notionary.page.content.markdown.nodes.base import MarkdownNode
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry
from notionary.page.content.syntax.definition.grammar import MarkdownGrammar


def flatten_children(children: list[MarkdownNode]) -> MarkdownNode | None:
    """Convert a list of child nodes to a single node for container compatibility."""
    if not children:
        return None
    if len(children) == 1:
        return children[0]
    return _MultiChildWrapper(children)


class _MultiChildWrapper(MarkdownNode):
    def __init__(self, children: list[MarkdownNode]) -> None:
        super().__init__()
        self.children = children

    def to_markdown(self) -> str:
        return "\n".join(child.to_markdown() for child in self.children if child)


class ContainerNode(MarkdownNode):
    children: list[MarkdownNode]

    def __init__(self, syntax_registry: SyntaxDefinitionRegistry | None = None) -> None:
        super().__init__(syntax_registry=syntax_registry)
        grammar = self._get_grammar(syntax_registry)
        self._spaces_per_nesting_level = grammar.spaces_per_nesting_level

    def render_children(self, indent_level: int = 1) -> str:
        if not self.children:
            return ""

        indent = self._calculate_indent(indent_level)
        rendered = [self._indent_child(child, indent) for child in self.children]
        rendered = [text for text in rendered if text]

        return f"\n{'\n'.join(rendered)}" if rendered else ""

    def render_child(self, child: MarkdownNode, indent_level: int = 1) -> str:
        child_markdown = child.to_markdown()
        if not child_markdown:
            return ""

        indent = self._calculate_indent(indent_level)
        return self._indent_text(child_markdown, indent)

    def _calculate_indent(self, level: int) -> str:
        return " " * (self._spaces_per_nesting_level * level)

    def _indent_child(self, child: MarkdownNode, indent: str) -> str:
        child_markdown = child.to_markdown()
        return self._indent_text(child_markdown, indent) if child_markdown else ""

    @staticmethod
    def _indent_text(text: str, indent: str) -> str:
        lines = text.split("\n")
        return "\n".join(f"{indent}{line}" if line.strip() else line for line in lines)

    @staticmethod
    def _get_grammar(
        syntax_registry: SyntaxDefinitionRegistry | None,
    ) -> MarkdownGrammar:
        return (
            syntax_registry._markdown_grammar if syntax_registry else MarkdownGrammar()
        )



from typing import override

from notionary.blocks.schemas import CreateDividerBlock, DividerData
from notionary.page.content.parser.parsers.base import (
    BlockParsingContext,
    LineParser,
)
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class DividerParser(LineParser):
    def __init__(self, syntax_registry: SyntaxDefinitionRegistry) -> None:
        super().__init__(syntax_registry)
        self._syntax = syntax_registry.get_divider_syntax()

    @override
    def _can_handle(self, context: BlockParsingContext) -> bool:
        if context.is_inside_parent_context():
            return False
        return self._is_divider(context.line)

    @override
    async def _process(self, context: BlockParsingContext) -> None:
        block = self._create_divider_block()
        if block:
            context.result_blocks.append(block)

    def _is_divider(self, line: str) -> bool:
        return self._syntax.regex_pattern.match(line) is not None

    def _create_divider_block(self) -> CreateDividerBlock:
        divider_data = DividerData()
        return CreateDividerBlock(divider=divider_data)



from typing import override

from notionary.blocks.rich_text.name_id_resolver.port import NameIdResolver
from notionary.workspace.query.service import WorkspaceQueryService


class DatabaseNameIdResolver(NameIdResolver):
    def __init__(self, search_service: WorkspaceQueryService | None = None) -> None:
        self.search_service = search_service or WorkspaceQueryService()

    @override
    async def resolve_name_to_id(self, name: str) -> str | None:
        if not name:
            return None

        cleaned_name = name.strip()
        database = await self.search_service.find_database(query=cleaned_name)
        return database.id if database else None

    @override
    async def resolve_id_to_name(self, database_id: str) -> str | None:
        if not database_id:
            return None

        try:
            from notionary import NotionDatabase

            database = await NotionDatabase.from_id(database_id)
            return database.title if database else None
        except Exception:
            return None



from notionary.exceptions.base import NotionaryException


class NotionApiError(NotionaryException):
    def __init__(
        self,
        message: str,
        status_code: int | None = None,
        response_text: int | None = None,
    ):
        super().__init__(message)
        self.status_code = status_code
        self.response_text = response_text


class NotionAuthenticationError(NotionApiError):
    pass


class NotionPermissionError(NotionApiError):
    pass


class NotionRateLimitError(NotionApiError):
    pass


class NotionResourceNotFoundError(NotionApiError):
    pass


class NotionValidationError(NotionApiError):
    pass


class NotionServerError(NotionApiError):
    pass


class NotionConnectionError(NotionApiError):
    pass



from collections.abc import Awaitable, Callable

from notionary.blocks.schemas import Block
from notionary.page.content.syntax.definition.grammar import MarkdownGrammar

ConvertChildrenCallback = Callable[[list[Block], int], Awaitable[str]]


class MarkdownRenderingContext:
    def __init__(
        self,
        block: Block,
        indent_level: int,
        convert_children_callback: ConvertChildrenCallback | None = None,
        markdown_grammar: MarkdownGrammar | None = None,
    ) -> None:
        self.block = block
        self.indent_level = indent_level
        self.convert_children_callback = convert_children_callback
        markdown_grammar = markdown_grammar or MarkdownGrammar()
        self._spaces_per_nesting_level = markdown_grammar.spaces_per_nesting_level

        self.markdown_result: str | None = None

    async def render_children(self) -> str:
        return await self._convert_children_to_markdown(self.indent_level)

    async def render_children_with_additional_indent(
        self, additional_indent: int
    ) -> str:
        return await self._convert_children_to_markdown(
            self.indent_level + additional_indent
        )

    async def _convert_children_to_markdown(self, indent_level: int) -> str:
        if not self._has_children() or not self.convert_children_callback:
            return ""

        return await self.convert_children_callback(
            self._get_children_blocks(), indent_level
        )

    def _get_children_blocks(self) -> list[Block]:
        if self._has_children():
            return self.block.children
        return []

    def _has_children(self) -> bool:
        return (
            self.block.has_children
            and self.block.children
            and len(self.block.children) > 0
        )

    def indent_text(self, text: str) -> str:
        if not text:
            return text

        spaces = " " * self._spaces_per_nesting_level * self.indent_level
        lines = text.split("\n")
        return "\n".join(f"{spaces}{line}" if line.strip() else line for line in lines)



from __future__ import annotations

from notionary.page.content.markdown.builder import MarkdownBuilder
from notionary.page.content.markdown.structured_output.models import (
    AudioSchema,
    BookmarkSchema,
    BreadcrumbSchema,
    BulletedListItemSchema,
    BulletedListSchema,
    CalloutSchema,
    CodeSchema,
    ColumnsSchema,
    EmbedSchema,
    EquationSchema,
    FileSchema,
    Heading1Schema,
    Heading2Schema,
    Heading3Schema,
    ImageSchema,
    MarkdownDocumentSchema,
    MarkdownNodeSchema,
    MermaidSchema,
    NumberedListItemSchema,
    NumberedListSchema,
    ParagraphSchema,
    PdfSchema,
    QuoteSchema,
    TableOfContentsSchema,
    TableSchema,
    TodoListSchema,
    TodoSchema,
    ToggleSchema,
    VideoSchema,
)
from notionary.utils.decorators import time_execution_sync
from notionary.utils.mixins.logging import LoggingMixin


class StructuredOutputMarkdownConverter(LoggingMixin):
    def __init__(self, builder: MarkdownBuilder | None = None) -> None:
        self.builder = builder or MarkdownBuilder()

    @time_execution_sync()
    def convert(self, schema: MarkdownDocumentSchema) -> str:
        for node in schema.nodes:
            self._process_node(node)
        return self.builder.build()

    def _process_node(self, node: MarkdownNodeSchema) -> None:
        node.process_with(self)

    def _process_heading_1(self, node: Heading1Schema) -> None:
        builder_func = (
            self._create_children_builder(node.children) if node.children else None
        )
        self.builder.h1(node.text, builder_func)

    def _process_heading_2(self, node: Heading2Schema) -> None:
        builder_func = (
            self._create_children_builder(node.children) if node.children else None
        )
        self.builder.h2(node.text, builder_func)

    def _process_heading_3(self, node: Heading3Schema) -> None:
        builder_func = (
            self._create_children_builder(node.children) if node.children else None
        )
        self.builder.h3(node.text, builder_func)

    def _process_paragraph(self, node: ParagraphSchema) -> None:
        self.builder.paragraph(node.text)

    def _process_space(self) -> None:
        self.builder.space()

    def _process_divider(self) -> None:
        self.builder.divider()

    def _process_quote(self, node: QuoteSchema) -> None:
        builder_func = (
            self._create_children_builder(node.children) if node.children else None
        )
        self.builder.quote(node.text, builder_func)

    def _process_bulleted_list(self, node: BulletedListSchema) -> None:
        has_children = any(item.children for item in node.items)

        if has_children:
            for item in node.items:
                self._process_bulleted_list_item(item)
        else:
            texts = [item.text for item in node.items]
            self.builder.bulleted_list(texts)

    def _process_bulleted_list_item(self, node: BulletedListItemSchema) -> None:
        builder_func = (
            self._create_children_builder(node.children) if node.children else None
        )
        self.builder.bulleted_list_item(node.text, builder_func)

    def _process_numbered_list(self, node: NumberedListSchema) -> None:
        has_children = any(item.children for item in node.items)

        if has_children:
            for item in node.items:
                self._process_numbered_list_item(item)
        else:
            texts = [item.text for item in node.items]
            self.builder.numbered_list(texts)

    def _process_numbered_list_item(self, node: NumberedListItemSchema) -> None:
        builder_func = (
            self._create_children_builder(node.children) if node.children else None
        )
        self.builder.numbered_list_item(node.text, builder_func)

    def _process_todo(self, node: TodoSchema) -> None:
        builder_func = (
            self._create_children_builder(node.children) if node.children else None
        )
        self.builder.todo(node.text, checked=node.checked, builder_func=builder_func)

    def _process_todo_list(self, node: TodoListSchema) -> None:
        has_children = any(item.children for item in node.items)

        if has_children:
            for todo_item in node.items:
                self._process_todo(todo_item)
        else:
            texts = [item.text for item in node.items]
            completed = [item.checked for item in node.items]
            self.builder.todo_list(texts, completed)

    def _process_callout(self, node: CalloutSchema) -> None:
        if node.children:
            builder_func = self._create_children_builder(node.children)
            self.builder.callout_with_children(node.text, node.emoji, builder_func)
        else:
            self.builder.callout(node.text, node.emoji)

    def _process_toggle(self, node: ToggleSchema) -> None:
        builder_func = self._create_children_builder(node.children)
        self.builder.toggle(node.title, builder_func)

    def _process_image(self, node: ImageSchema) -> None:
        self.builder.image(node.url, node.caption)

    def _process_video(self, node: VideoSchema) -> None:
        self.builder.video(node.url, node.caption)

    def _process_audio(self, node: AudioSchema) -> None:
        self.builder.audio(node.url, node.caption)

    def _process_file(self, node: FileSchema) -> None:
        self.builder.file(node.url, node.caption)

    def _process_pdf(self, node: PdfSchema) -> None:
        self.builder.pdf(node.url, node.caption)

    def _process_bookmark(self, node: BookmarkSchema) -> None:
        self.builder.bookmark(node.url, node.title, node.caption)

    def _process_embed(self, node: EmbedSchema) -> None:
        self.builder.embed(node.url, node.caption)

    def _process_code(self, node: CodeSchema) -> None:
        self.builder.code(node.code, node.language, node.caption)

    def _process_mermaid(self, node: MermaidSchema) -> None:
        self.builder.mermaid(node.diagram, node.caption)

    def _process_table(self, node: TableSchema) -> None:
        self.builder.table(node.headers, node.rows)

    def _process_breadcrumb(self, node: BreadcrumbSchema) -> None:
        self.builder.breadcrumb()

    def _process_equation(self, node: EquationSchema) -> None:
        self.builder.equation(node.expression)

    def _process_table_of_contents(self, node: TableOfContentsSchema) -> None:
        self.builder.table_of_contents()

    def _process_columns(self, node: ColumnsSchema) -> None:
        builder_funcs = []
        width_ratios = []

        for column in node.columns:
            builder_func = self._create_children_builder(column.children)
            builder_funcs.append(builder_func)
            width_ratios.append(column.width_ratio)

        if any(r is not None for r in width_ratios):
            self.builder.columns(*builder_funcs, width_ratios=width_ratios)
        else:
            self.builder.columns(*builder_funcs)

    def _create_children_builder(self, children: list[MarkdownNodeSchema] | None):
        if not children:
            return None

        captured_children = children

        def builder_func(builder: MarkdownBuilder) -> MarkdownBuilder:
            converter = StructuredOutputMarkdownConverter()
            converter.builder = builder
            for child in captured_children:
                converter._process_node(child)
            return builder

        return builder_func



from typing import override

from notionary.blocks.schemas import CreateImageBlock, ExternalFileWithCaption
from notionary.page.content.parser.parsers.file_like_block import FileLikeBlockParser
from notionary.page.content.syntax.definition import (
    SyntaxDefinition,
    SyntaxDefinitionRegistry,
)


class ImageParser(FileLikeBlockParser[CreateImageBlock]):
    @override
    def _get_syntax(
        self, syntax_registry: SyntaxDefinitionRegistry
    ) -> SyntaxDefinition:
        return syntax_registry.get_image_syntax()

    @override
    def _create_block(self, file_data: ExternalFileWithCaption) -> CreateImageBlock:
        return CreateImageBlock(image=file_data)



from .models import (
    AnyMarkdownNode,
    AudioSchema,
    BookmarkSchema,
    BreadcrumbSchema,
    BulletedListItemSchema,
    BulletedListSchema,
    CalloutSchema,
    CodeSchema,
    ColumnSchema,
    ColumnsSchema,
    DividerSchema,
    EmbedSchema,
    EquationSchema,
    FileSchema,
    Heading1Schema,
    Heading2Schema,
    Heading3Schema,
    ImageSchema,
    MarkdownDocumentSchema,
    MarkdownNodeSchema,
    MermaidSchema,
    NumberedListItemSchema,
    NumberedListSchema,
    ParagraphSchema,
    PdfSchema,
    QuoteSchema,
    SpaceSchema,
    TableOfContentsSchema,
    TableSchema,
    TodoListSchema,
    TodoSchema,
    ToggleSchema,
    VideoSchema,
)
from .service import StructuredOutputMarkdownConverter

__all__ = [
    "AnyMarkdownNode",
    "AudioSchema",
    "BookmarkSchema",
    "BreadcrumbSchema",
    "BulletedListItemSchema",
    "BulletedListSchema",
    "CalloutSchema",
    "CodeSchema",
    "ColumnSchema",
    "ColumnsSchema",
    "DividerSchema",
    "EmbedSchema",
    "EquationSchema",
    "FileSchema",
    "Heading1Schema",
    "Heading2Schema",
    "Heading3Schema",
    "ImageSchema",
    "MarkdownDocumentSchema",
    "MarkdownNodeSchema",
    "MermaidSchema",
    "NumberedListItemSchema",
    "NumberedListSchema",
    "ParagraphSchema",
    "PdfSchema",
    "QuoteSchema",
    "SpaceSchema",
    "StructuredOutputMarkdownConverter",
    "TableOfContentsSchema",
    "TableSchema",
    "TodoListSchema",
    "TodoSchema",
    "ToggleSchema",
    "VideoSchema",
]



from typing import override

from notionary.blocks.rich_text.markdown_rich_text_converter import (
    MarkdownRichTextConverter,
)
from notionary.blocks.schemas import BlockColor, CreateQuoteBlock, CreateQuoteData
from notionary.page.content.parser.parsers.base import (
    BlockParsingContext,
    LineParser,
)
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class QuoteParser(LineParser):
    def __init__(
        self,
        syntax_registry: SyntaxDefinitionRegistry,
        rich_text_converter: MarkdownRichTextConverter,
    ) -> None:
        super().__init__(syntax_registry)
        self._syntax = syntax_registry.get_quote_syntax()
        self._rich_text_converter = rich_text_converter

    @override
    def _can_handle(self, context: BlockParsingContext) -> bool:
        if context.is_inside_parent_context():
            return False
        return self._is_quote(context.line)

    def _is_quote(self, line: str) -> bool:
        return self._syntax.regex_pattern.match(line) is not None

    @override
    async def _process(self, context: BlockParsingContext) -> None:
        quote_lines = self._collect_quote_lines(context)

        block = await self._create_quote_block(quote_lines)
        if not block:
            return

        # Lines consumed: all quote lines minus the current line (which is already being processed)
        context.lines_consumed = len(quote_lines) - 1

        await self._process_nested_children(block, context, quote_lines)
        context.result_blocks.append(block)

    def _collect_quote_lines(self, context: BlockParsingContext) -> list[str]:
        quote_lines = [context.line]
        for line in context.get_remaining_lines():
            if not self._is_quote(line):
                break
            quote_lines.append(line)
        return quote_lines

    async def _process_nested_children(
        self,
        block: CreateQuoteBlock,
        context: BlockParsingContext,
        quote_lines: list[str],
    ) -> None:
        # Calculate indent level after all quote lines
        last_quote_line_index = len(quote_lines) - 1
        child_lines = self._collect_child_lines_after_quote(
            context, last_quote_line_index
        )

        if not child_lines:
            return

        child_blocks = await self._parse_child_blocks(child_lines, context)
        if child_blocks:
            block.quote.children = child_blocks

        context.lines_consumed += len(child_lines)

    def _collect_child_lines_after_quote(
        self, context: BlockParsingContext, last_quote_index: int
    ) -> list[str]:
        """Collect indented children after the quote block."""
        parent_indent_level = context.get_line_indentation_level()
        remaining_lines = context.get_remaining_lines()

        # Skip the quote lines we already processed
        lines_after_quote = remaining_lines[last_quote_index:]

        child_lines = []
        expected_child_indent = parent_indent_level + 1

        for line in lines_after_quote:
            if not line.strip():
                child_lines.append(line)
                continue

            line_indent = context.get_line_indentation_level(line)
            if line_indent >= expected_child_indent:
                child_lines.append(line)
            else:
                break

        return child_lines

    async def _parse_child_blocks(
        self, child_lines: list[str], context: BlockParsingContext
    ) -> list[CreateQuoteBlock]:
        stripped_lines = self._remove_parent_indentation(child_lines, context)
        children_text = self._convert_lines_to_text(stripped_lines)
        return await context.parse_nested_markdown(children_text)

    def _remove_parent_indentation(
        self, lines: list[str], context: BlockParsingContext
    ) -> list[str]:
        return context.strip_indentation_level(lines, levels=1)

    def _convert_lines_to_text(self, lines: list[str]) -> str:
        return "\n".join(lines)

    async def _create_quote_block(
        self, quote_lines: list[str]
    ) -> CreateQuoteBlock | None:
        contents = self._extract_quote_contents(quote_lines)
        if not contents:
            return None

        content = self._join_contents_for_multiline_quote(contents)
        rich_text = await self._convert_to_rich_text(content)
        return self._build_block(rich_text)

    def _extract_quote_contents(self, quote_lines: list[str]) -> list[str]:
        contents = []
        for line in quote_lines:
            match = self._syntax.regex_pattern.match(line)
            if match:
                contents.append(match.group(1).strip())
        return contents

    def _join_contents_for_multiline_quote(self, contents: list[str]) -> str:
        return "\n".join(contents)

    async def _convert_to_rich_text(self, content: str):
        return await self._rich_text_converter.to_rich_text(content)

    def _build_block(self, rich_text) -> CreateQuoteBlock:
        quote_data = CreateQuoteData(rich_text=rich_text, color=BlockColor.DEFAULT)
        return CreateQuoteBlock(quote=quote_data)



from .factory import PageContentServiceFactory
from .service import PageContentService
from .syntax.prompts import SyntaxPromptRegistry

__all__ = [
    "PageContentService",
    "PageContentServiceFactory",
    "SyntaxPromptRegistry",
]



import re


def extract_uuid(source: str) -> str | None:
    UUID_RAW_PATTERN = r"([a-f0-9]{32})"

    if _is_valid_uuid(source):
        return source

    match = re.search(UUID_RAW_PATTERN, source.lower())
    if not match:
        return None

    uuid_raw = match.group(1)
    return f"{uuid_raw[0:8]}-{uuid_raw[8:12]}-{uuid_raw[12:16]}-{uuid_raw[16:20]}-{uuid_raw[20:32]}"


def _is_valid_uuid(uuid: str) -> bool:
    UUID_PATTERN = r"^[a-f0-9]{8}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{4}-[a-f0-9]{12}$"
    return bool(re.match(UUID_PATTERN, uuid.lower()))



import asyncio
import functools
import logging
import time
from collections.abc import Callable, Coroutine
from typing import Any, ParamSpec, TypeVar

P = ParamSpec("P")
R = TypeVar("R")
T = TypeVar("T")

type _SyncFunc = Callable[P, R]
type _AsyncFunc = Callable[P, Coroutine[Any, Any, R]]
type _SyncDecorator = Callable[[_SyncFunc], _SyncFunc]
type _AsyncDecorator = Callable[[_AsyncFunc], _AsyncFunc]


def singleton(cls):
    instance = [None]

    def wrapper(*args, **kwargs):
        if instance[0] is None:
            instance[0] = cls(*args, **kwargs)
        return instance[0]

    return wrapper


def time_execution_sync(
    additional_text: str = "", min_duration_to_log: float = 0.25
) -> _SyncDecorator:
    def decorator(func: _SyncFunc) -> _SyncFunc:
        @functools.wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            start_time = time.perf_counter()
            result = func(*args, **kwargs)
            execution_time = time.perf_counter() - start_time

            if execution_time > min_duration_to_log:
                logger = _get_logger_from_context(args, func)
                function_name = additional_text.strip("-") or func.__name__
                logger.debug(f"⏳ {function_name}() took {execution_time:.2f}s")

            return result

        return wrapper

    return decorator


def time_execution_async(
    additional_text: str = "",
    min_duration_to_log: float = 0.25,
) -> _AsyncDecorator:
    def decorator(func: _AsyncFunc) -> _AsyncFunc:
        @functools.wraps(func)
        async def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
            start_time = time.perf_counter()
            result = await func(*args, **kwargs)
            execution_time = time.perf_counter() - start_time

            if execution_time > min_duration_to_log:
                logger = _get_logger_from_context(args, func)
                function_name = additional_text.strip("-") or func.__name__
                logger.debug(f"⏳ {function_name}() took {execution_time:.2f}s")

            return result

        return wrapper

    return decorator


def _get_logger_from_context(args: tuple, func: Callable) -> logging.Logger:
    if _has_instance_logger(args):
        return _extract_instance_logger(args)

    return _get_module_logger(func)


def _has_instance_logger(args: tuple) -> bool:
    return bool(args) and hasattr(args[0], "logger")


def _extract_instance_logger(args: tuple) -> logging.Logger:
    return args[0].logger


def _get_module_logger(func: Callable) -> logging.Logger:
    return logging.getLogger(func.__module__)


def async_retry(
    max_retries: int = 3,
    initial_delay: float = 1.0,
    backoff_factor: float = 2.0,
    retry_on_exceptions: tuple[type[Exception], ...] | None = None,
):
    def decorator(func: Callable[..., Any]) -> Callable[..., Any]:
        @functools.wraps(func)
        async def wrapper(*args: Any, **kwargs: Any) -> Any:
            delay = initial_delay
            last_exception = None

            for attempt in range(max_retries + 1):
                try:
                    return await func(*args, **kwargs)
                except Exception as e:
                    last_exception = e

                    if retry_on_exceptions is not None and not isinstance(
                        e, retry_on_exceptions
                    ):
                        raise

                    if attempt == max_retries:
                        raise

                    await asyncio.sleep(delay)
                    delay *= backoff_factor

            raise last_exception

        return wrapper

    return decorator



from typing import override

from notionary.page.content.markdown.nodes.base import MarkdownNode
from notionary.page.content.markdown.nodes.mixins.caption import (
    CaptionMarkdownNodeMixin,
)
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class VideoMarkdownNode(MarkdownNode, CaptionMarkdownNodeMixin):
    def __init__(
        self,
        url: str,
        caption: str | None = None,
        syntax_registry: SyntaxDefinitionRegistry | None = None,
    ) -> None:
        super().__init__(syntax_registry=syntax_registry)
        self.url = url
        self.caption = caption

    @override
    def to_markdown(self) -> str:
        video_syntax = self._syntax_registry.get_video_syntax()
        base_markdown = (
            f"{video_syntax.start_delimiter}{self.url}{video_syntax.end_delimiter}"
        )
        return self._append_caption_to_markdown(base_markdown, self.caption)



from typing import override

from notionary.blocks.rich_text.rich_text_markdown_converter import (
    RichTextToMarkdownConverter,
)
from notionary.blocks.schemas import Block, BlockType
from notionary.page.content.renderer.context import MarkdownRenderingContext
from notionary.page.content.renderer.renderers.base import BlockRenderer
from notionary.page.content.syntax.definition import (
    MarkdownGrammar,
    SyntaxDefinitionRegistry,
)


class NumberedListRenderer(BlockRenderer):
    def __init__(
        self,
        syntax_registry: SyntaxDefinitionRegistry | None = None,
        rich_text_markdown_converter: RichTextToMarkdownConverter | None = None,
        markdown_grammar: MarkdownGrammar | None = None,
    ) -> None:
        super().__init__(syntax_registry=syntax_registry)
        self._rich_text_markdown_converter = (
            rich_text_markdown_converter or RichTextToMarkdownConverter()
        )

        markdown_grammar = markdown_grammar or MarkdownGrammar()
        self._numbered_list_placeholder = markdown_grammar.numbered_list_placeholder

    @override
    def _can_handle(self, block: Block) -> bool:
        return block.type == BlockType.NUMBERED_LIST_ITEM

    @override
    async def _process(self, context: MarkdownRenderingContext) -> None:
        list_item_data = context.block.numbered_list_item
        rich_text = list_item_data.rich_text if list_item_data else []
        content = await self._rich_text_markdown_converter.to_markdown(rich_text)

        item_line = context.indent_text(f"{self._numbered_list_placeholder}. {content}")

        children_markdown = await context.render_children_with_additional_indent(1)

        if children_markdown:
            context.markdown_result = f"{item_line}\n{children_markdown}"
        else:
            context.markdown_result = item_line



from typing import override

from notionary.page.content.markdown.nodes.base import MarkdownNode
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class TableMarkdownNode(MarkdownNode):
    def __init__(
        self,
        headers: list[str],
        rows: list[list[str]],
        syntax_registry: SyntaxDefinitionRegistry | None = None,
    ) -> None:
        super().__init__(syntax_registry=syntax_registry)
        self._validate_input(headers, rows)
        self.headers = headers
        self.rows = rows

    def _validate_input(self, headers: list[str], rows: list[list[str]]) -> None:
        if not headers:
            raise ValueError("headers must not be empty")
        if not all(isinstance(row, list) for row in rows):
            raise ValueError("rows must be a list of lists")

    @override
    def to_markdown(self) -> str:
        header = self._build_header_row()
        separator = self._build_separator_row()
        data_rows = self._build_data_rows()
        return "\n".join([header, separator, *data_rows])

    def _build_header_row(self) -> str:
        return self._format_row(self.headers)

    def _format_row(self, cells: list[str]) -> str:
        table_syntax = self._syntax_registry.get_table_syntax()
        delimiter = table_syntax.start_delimiter
        joined_cells = f" {delimiter} ".join(cells)
        return f"{delimiter} {joined_cells} {delimiter}"

    def _build_separator_row(self) -> str:
        col_count = len(self.headers)
        separators = ["-"] * col_count
        row = self._format_row(separators)
        return row

    def _build_data_rows(self) -> list[str]:
        return [self._format_row(row) for row in self.rows]



from typing import override

from notionary.blocks.rich_text.name_id_resolver.port import NameIdResolver
from notionary.workspace.query.service import WorkspaceQueryService


class PageNameIdResolver(NameIdResolver):
    def __init__(self, search_service: WorkspaceQueryService | None = None) -> None:
        self.search_service = search_service or WorkspaceQueryService()

    @override
    async def resolve_name_to_id(self, name: str) -> str | None:
        if not name:
            return None

        cleaned_name = name.strip()
        return await self._resolve_page_id(cleaned_name)

    @override
    async def resolve_id_to_name(self, page_id: str) -> str | None:
        if not page_id:
            return None

        try:
            from notionary import NotionPage

            page = await NotionPage.from_id(page_id)
            return page.title if page else None
        except Exception:
            return None

    async def _resolve_page_id(self, name: str) -> str | None:
        page = await self.search_service.find_page(query=name)
        return page.id if page else None



from .column_syntax import ColumnSyntaxPreProcessor
from .indentation import IndentationNormalizer
from .port import PreProcessor
from .video_syntax import VideoFormatPreProcessor
from .whitespace import WhitespacePreProcessor

__all__ = [
    "ColumnSyntaxPreProcessor",
    "IndentationNormalizer",
    "PreProcessor",
    "VideoFormatPreProcessor",
    "WhitespacePreProcessor",
]



from notionary.exceptions.base import NotionaryException


class UnsupportedFileTypeException(NotionaryException):
    def __init__(
        self,
        extension: str,
        filename: str,
        supported_extensions_by_category: dict[str, list[str]],
    ):
        supported_exts = []
        for category, extensions in supported_extensions_by_category.items():
            supported_exts.append(f"{category}: {', '.join(extensions[:5])}...")

        supported_info = "\n  ".join(supported_exts)
        super().__init__(
            f"File '{filename}' has unsupported extension '{extension}'.\n"
            f"Supported file types by category:\n  {supported_info}"
        )
        self.extension = extension
        self.filename = filename


class NoFileExtensionException(NotionaryException):
    def __init__(self, filename: str):
        super().__init__(
            f"File '{filename}' has no extension. Files must have a valid extension to determine their type."
        )
        self.filename = filename


class FileSizeException(NotionaryException):
    def __init__(self, filename: str, file_size_bytes: int, max_size_bytes: int):
        file_size_mb = file_size_bytes / (1024 * 1024)
        max_size_mb = max_size_bytes / (1024 * 1024)
        super().__init__(
            f"File '{filename}' is too large ({file_size_mb:.2f} MB). Maximum allowed size is {max_size_mb:.2f} MB."
        )
        self.filename = filename
        self.file_size_bytes = file_size_bytes
        self.max_size_bytes = max_size_bytes


class FileNotFoundError(NotionaryException):
    def __init__(self, file_path: str):
        super().__init__(f"File does not exist: {file_path}")
        self.file_path = file_path


class FilenameTooLongError(NotionaryException):
    def __init__(self, filename: str, filename_bytes: int, max_filename_bytes: int):
        super().__init__(
            f"Filename too long: {filename_bytes} bytes (max {max_filename_bytes}). Filename: {filename}"
        )
        self.filename = filename
        self.filename_bytes = filename_bytes
        self.max_filename_bytes = max_filename_bytes


class UploadFailedError(NotionaryException):
    def __init__(self, file_upload_id: str, reason: str | None = None):
        message = f"Upload failed for file_upload_id: {file_upload_id}"
        if reason:
            message += f". Reason: {reason}"
        super().__init__(message)
        self.file_upload_id = file_upload_id
        self.reason = reason


class UploadTimeoutError(NotionaryException):
    def __init__(self, file_upload_id: str, timeout_seconds: int):
        super().__init__(
            f"Upload timeout after {timeout_seconds}s for file_upload_id: {file_upload_id}"
        )
        self.file_upload_id = file_upload_id
        self.timeout_seconds = timeout_seconds



from enum import StrEnum
from typing import Literal

from pydantic import BaseModel


class ParentType(StrEnum):
    DATABASE_ID = "database_id"
    DATA_SOURCE_ID = "data_source_id"
    PAGE_ID = "page_id"
    BLOCK_ID = "block_id"
    WORKSPACE = "workspace"


class DataSourceParent(BaseModel):
    type: Literal[ParentType.DATA_SOURCE_ID]
    data_source_id: str
    database_id: str


class PageParent(BaseModel):
    type: Literal[ParentType.PAGE_ID]
    page_id: str


class BlockParent(BaseModel):
    type: Literal[ParentType.BLOCK_ID]
    block_id: str


class WorkspaceParent(BaseModel):
    type: Literal[ParentType.WORKSPACE]
    workspace: bool


class DatabaseParent(BaseModel):
    type: Literal[ParentType.DATABASE_ID]
    database_id: str


Parent = DataSourceParent | PageParent | BlockParent | WorkspaceParent | DatabaseParent



from typing import override

from notionary.page.content.markdown.nodes.base import MarkdownNode
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class BreadcrumbMarkdownNode(MarkdownNode):
    def __init__(self, syntax_registry: SyntaxDefinitionRegistry | None = None) -> None:
        super().__init__(syntax_registry=syntax_registry)

    @override
    def to_markdown(self) -> str:
        breadcrumb_syntax = self._syntax_registry.get_breadcrumb_syntax()
        return breadcrumb_syntax.start_delimiter



from enum import StrEnum


class AudioExtension(StrEnum):
    AAC = ".aac"
    ADTS = ".adts"
    MID = ".mid"
    MIDI = ".midi"
    MP3 = ".mp3"
    MPGA = ".mpga"
    M4A = ".m4a"
    M4B = ".m4b"
    MP4 = ".mp4"
    OGA = ".oga"
    OGG = ".ogg"
    WAV = ".wav"
    WMA = ".wma"


class AudioMimeType(StrEnum):
    AAC = "audio/aac"
    MIDI = "audio/midi"
    MPEG = "audio/mpeg"
    MP4 = "audio/mp4"
    OGG = "audio/ogg"
    WAV = "audio/wav"
    WMA = "audio/x-ms-wma"


class DocumentExtension(StrEnum):
    PDF = ".pdf"
    TXT = ".txt"
    JSON = ".json"
    DOC = ".doc"
    DOT = ".dot"
    DOCX = ".docx"
    DOTX = ".dotx"
    XLS = ".xls"
    XLT = ".xlt"
    XLA = ".xla"
    XLSX = ".xlsx"
    XLTX = ".xltx"
    PPT = ".ppt"
    POT = ".pot"
    PPS = ".pps"
    PPA = ".ppa"
    PPTX = ".pptx"
    POTX = ".potx"


class DocumentMimeType(StrEnum):
    PDF = "application/pdf"
    PLAIN_TEXT = "text/plain"
    JSON = "application/json"
    MSWORD = "application/msword"
    WORD_DOCUMENT = (
        "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
    )
    WORD_TEMPLATE = (
        "application/vnd.openxmlformats-officedocument.wordprocessingml.template"
    )
    EXCEL = "application/vnd.ms-excel"
    EXCEL_SHEET = "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet"
    EXCEL_TEMPLATE = (
        "application/vnd.openxmlformats-officedocument.spreadsheetml.template"
    )
    POWERPOINT = "application/vnd.ms-powerpoint"
    POWERPOINT_PRESENTATION = (
        "application/vnd.openxmlformats-officedocument.presentationml.presentation"
    )
    POWERPOINT_TEMPLATE = (
        "application/vnd.openxmlformats-officedocument.presentationml.template"
    )


class ImageExtension(StrEnum):
    GIF = ".gif"
    HEIC = ".heic"
    JPEG = ".jpeg"
    JPG = ".jpg"
    PNG = ".png"
    SVG = ".svg"
    TIF = ".tif"
    TIFF = ".tiff"
    WEBP = ".webp"
    ICO = ".ico"


class ImageMimeType(StrEnum):
    GIF = "image/gif"
    HEIC = "image/heic"
    JPEG = "image/jpeg"
    PNG = "image/png"
    SVG = "image/svg+xml"
    TIFF = "image/tiff"
    WEBP = "image/webp"
    ICON = "image/vnd.microsoft.icon"


class VideoExtension(StrEnum):
    AMV = ".amv"
    ASF = ".asf"
    WMV = ".wmv"
    AVI = ".avi"
    F4V = ".f4v"
    FLV = ".flv"
    GIFV = ".gifv"
    M4V = ".m4v"
    MP4 = ".mp4"
    MKV = ".mkv"
    WEBM = ".webm"
    MOV = ".mov"
    QT = ".qt"
    MPEG = ".mpeg"


class VideoMimeType(StrEnum):
    AMV = "video/x-amv"
    ASF = "video/x-ms-asf"
    AVI = "video/x-msvideo"
    F4V = "video/x-f4v"
    FLV = "video/x-flv"
    MP4 = "video/mp4"
    MKV = "video/x-matroska"
    WEBM = "video/webm"
    QUICKTIME = "video/quicktime"
    MPEG = "video/mpeg"


class FileCategory(StrEnum):
    AUDIO = "audio"
    DOCUMENT = "document"
    IMAGE = "image"
    VIDEO = "video"



from .config import FileUploadConfig
from .constants import (
    NOTION_MAX_FILENAME_BYTES,
    NOTION_MULTI_PART_CHUNK_SIZE_MAX,
    NOTION_MULTI_PART_CHUNK_SIZE_MIN,
    NOTION_RECOMMENDED_CHUNK_SIZE,
    NOTION_SINGLE_PART_MAX_SIZE,
)

__all__ = [
    "NOTION_MAX_FILENAME_BYTES",
    "NOTION_MULTI_PART_CHUNK_SIZE_MAX",
    "NOTION_MULTI_PART_CHUNK_SIZE_MIN",
    "NOTION_RECOMMENDED_CHUNK_SIZE",
    "NOTION_SINGLE_PART_MAX_SIZE",
    "FileUploadConfig",
]



from pydantic import TypeAdapter

from notionary.http.client import NotionHttpClient
from notionary.user.schemas import (
    BotUserResponseDto,
    NotionUsersListResponse,
    UserResponseDto,
)
from notionary.utils.pagination import paginate_notion_api


class UserHttpClient(NotionHttpClient):
    async def get_user_by_id(self, user_id: str) -> UserResponseDto:
        response = await self.get(f"users/{user_id}")

        adapter = TypeAdapter(UserResponseDto)
        return adapter.validate_python(response)

    async def get_all_workspace_users(self) -> list[UserResponseDto]:
        all_entities = await paginate_notion_api(
            self._get_workspace_entities, page_size=100
        )

        self.logger.info("Fetched %d total workspace users", len(all_entities))
        return all_entities

    async def _get_workspace_entities(
        self, page_size: int = 100, start_cursor: str | None = None
    ) -> NotionUsersListResponse | None:
        params = {"page_size": min(page_size, 100)}
        if start_cursor:
            params["start_cursor"] = start_cursor

        response = await self.get("users", params=params)

        return NotionUsersListResponse.model_validate(response)

    async def get_current_integration_bot(self) -> BotUserResponseDto:
        response = await self.get("users/me")

        return BotUserResponseDto.model_validate(response)



from .data_source.service import NotionDataSource
from .database.service import NotionDatabase
from .file_upload import FileUploadQuery, FileUploadQueryBuilder, NotionFileUpload
from .page.content import SyntaxPromptRegistry
from .page.content.markdown.builder import MarkdownBuilder
from .page.content.markdown.structured_output import (
    MarkdownDocumentSchema,
    StructuredOutputMarkdownConverter,
)
from .page.service import NotionPage
from .workspace import (
    NotionWorkspace,
    NotionWorkspaceQueryConfigBuilder,
    WorkspaceQueryConfig,
)

__all__ = [
    "AudioSchema",
    "BookmarkSchema",
    "BreadcrumbSchema",
    "BulletedListItemSchema",
    "BulletedListSchema",
    "CalloutSchema",
    "CodeSchema",
    "ColumnSchema",
    "ColumnsSchema",
    "DividerSchema",
    "EmbedSchema",
    "EquationSchema",
    "FileSchema",
    "FileUploadQuery",
    "FileUploadQueryBuilder",
    "Heading1Schema",
    "Heading2Schema",
    "Heading3Schema",
    "ImageSchema",
    "MarkdownBuilder",
    "MarkdownDocumentSchema",
    "MarkdownNodeSchema",
    "MermaidSchema",
    "NotionDataSource",
    "NotionDatabase",
    "NotionFileUpload",
    "NotionPage",
    "NotionWorkspace",
    "NotionWorkspaceQueryConfigBuilder",
    "NumberedListItemSchema",
    "NumberedListSchema",
    "ParagraphSchema",
    "PdfSchema",
    "QuoteSchema",
    "SpaceSchema",
    "StructuredOutputMarkdownConverter",
    "SyntaxPromptRegistry",
    "TableOfContentsSchema",
    "TableSchema",
    "TodoListSchema",
    "TodoSchema",
    "ToggleSchema",
    "VideoSchema",
    "WorkspaceQueryConfig",
]



from typing import override

from notionary.blocks.enums import BlockType
from notionary.blocks.rich_text.rich_text_markdown_converter import (
    RichTextToMarkdownConverter,
)
from notionary.blocks.schemas import Block
from notionary.page.content.renderer.context import MarkdownRenderingContext
from notionary.page.content.renderer.renderers.base import BlockRenderer
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class ToggleRenderer(BlockRenderer):
    def __init__(
        self,
        syntax_registry: SyntaxDefinitionRegistry | None = None,
        rich_text_markdown_converter: RichTextToMarkdownConverter | None = None,
    ) -> None:
        super().__init__(syntax_registry=syntax_registry)
        self._rich_text_markdown_converter = (
            rich_text_markdown_converter or RichTextToMarkdownConverter()
        )

    @override
    def _can_handle(self, block: Block) -> bool:
        return block.type == BlockType.TOGGLE

    @override
    async def _process(self, context: MarkdownRenderingContext) -> None:
        toggle_title = await self._extract_toggle_title(context.block)

        if not toggle_title:
            return

        syntax = self._syntax_registry.get_toggle_syntax()
        toggle_start = f"{syntax.start_delimiter} {toggle_title}"

        if context.indent_level > 0:
            toggle_start = context.indent_text(toggle_start)

        original_indent = context.indent_level
        context.indent_level += 1
        children_markdown = await context.render_children()
        context.indent_level = original_indent

        if children_markdown:
            context.markdown_result = f"{toggle_start}\n{children_markdown}"
        else:
            context.markdown_result = toggle_start

    async def _extract_toggle_title(self, block: Block) -> str:
        if not block.toggle or not block.toggle.rich_text:
            return ""

        rich_text_title = block.toggle.rich_text
        return await self._rich_text_markdown_converter.to_markdown(rich_text_title)



from typing import override

from notionary.blocks.rich_text.markdown_rich_text_converter import (
    MarkdownRichTextConverter,
)
from notionary.blocks.schemas import BlockColor, CreateToggleBlock, CreateToggleData
from notionary.page.content.parser.parsers import (
    BlockParsingContext,
    LineParser,
)
from notionary.page.content.syntax.definition import SyntaxDefinitionRegistry


class ToggleParser(LineParser):
    def __init__(
        self,
        syntax_registry: SyntaxDefinitionRegistry,
        rich_text_converter: MarkdownRichTextConverter,
    ) -> None:
        super().__init__(syntax_registry)
        self._syntax = syntax_registry.get_toggle_syntax()
        self._heading_syntax = syntax_registry.get_toggleable_heading_syntax()
        self._rich_text_converter = rich_text_converter

    @override
    def _can_handle(self, context: BlockParsingContext) -> bool:
        return self._is_toggle_start(context)

    @override
    async def _process(self, context: BlockParsingContext) -> None:
        if self._is_toggle_start(context):
            await self._process_toggle(context)

    def _is_toggle_start(self, context: BlockParsingContext) -> bool:
        if not self._syntax.regex_pattern.match(context.line):
            return False

        # Exclude toggleable heading patterns to be more resilient to wrong order of chain
        return not self.is_heading_start(context.line)

    def is_heading_start(self, line: str) -> bool:
        return self._heading_syntax.regex_pattern.match(line) is not None

    async def _process_toggle(self, context: BlockParsingContext) -> None:
        block = await self._create_toggle_block(context.line)
        if not block:
            return

        await self._process_nested_children(block, context)

        context.result_blocks.append(block)

    async def _create_toggle_block(self, line: str) -> CreateToggleBlock | None:
        if not (match := self._syntax.regex_pattern.match(line)):
            return None

        title = match.group(1).strip()
        rich_text = await self._rich_text_converter.to_rich_text(title)

        toggle_content = CreateToggleData(
            rich_text=rich_text, color=BlockColor.DEFAULT, children=[]
        )
        return CreateToggleBlock(toggle=toggle_content)

    async def _process_nested_children(
        self, block: CreateToggleBlock, context: BlockParsingContext
    ) -> None:
        parent_indent_level = context.get_line_indentation_level()
        child_lines = context.collect_indented_child_lines(parent_indent_level)

        if not child_lines:
            return

        stripped_lines = context.strip_indentation_level(child_lines, levels=1)
        child_markdown = "\n".join(stripped_lines)

        child_blocks = await context.parse_nested_markdown(child_markdown)
        block.toggle.children = child_blocks

        context.lines_consumed = len(child_lines)



import re

from notionary.page.content.syntax.definition.grammar import MarkdownGrammar
from notionary.page.content.syntax.definition.models import (
    EnclosedSyntaxDefinition,
    SimpleSyntaxDefinition,
    SyntaxDefinition,
    SyntaxDefinitionRegistryKey,
)


class SyntaxDefinitionRegistry:
    def __init__(
        self, markdown_markdown_grammar: MarkdownGrammar | None = None
    ) -> None:
        self._markdown_grammar = markdown_markdown_grammar or MarkdownGrammar()
        self._definitions: dict[SyntaxDefinitionRegistryKey, SyntaxDefinition] = {}
        self._register_defaults()

    def get_breadcrumb_syntax(self) -> SimpleSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.BREADCRUMB]

    def get_bulleted_list_syntax(self) -> SimpleSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.BULLETED_LIST]

    def get_divider_syntax(self) -> SimpleSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.DIVIDER]

    def get_numbered_list_syntax(self) -> SimpleSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.NUMBERED_LIST]

    def get_quote_syntax(self) -> SimpleSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.QUOTE]

    def get_table_syntax(self) -> SimpleSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.TABLE]

    def get_table_row_syntax(self) -> SimpleSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.TABLE_ROW]

    def get_table_of_contents_syntax(self) -> SimpleSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.TABLE_OF_CONTENTS]

    def get_todo_syntax(self) -> SimpleSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.TO_DO]

    def get_todo_done_syntax(self) -> SimpleSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.TO_DO_DONE]

    def get_caption_syntax(self) -> SimpleSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.CAPTION]

    def get_space_syntax(self) -> SimpleSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.SPACE]

    def get_heading_syntax(self) -> SimpleSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.HEADING]

    def get_audio_syntax(self) -> EnclosedSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.AUDIO]

    def get_bookmark_syntax(self) -> EnclosedSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.BOOKMARK]

    def get_embed_syntax(self) -> EnclosedSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.EMBED]

    def get_file_syntax(self) -> EnclosedSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.FILE]

    def get_image_syntax(self) -> EnclosedSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.IMAGE]

    def get_pdf_syntax(self) -> EnclosedSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.PDF]

    def get_video_syntax(self) -> EnclosedSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.VIDEO]

    def get_callout_syntax(self) -> EnclosedSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.CALLOUT]

    def get_code_syntax(self) -> EnclosedSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.CODE]

    def get_column_syntax(self) -> EnclosedSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.COLUMN]

    def get_column_list_syntax(self) -> EnclosedSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.COLUMN_LIST]

    def get_equation_syntax(self) -> EnclosedSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.EQUATION]

    def get_toggle_syntax(self) -> EnclosedSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.TOGGLE]

    def get_toggleable_heading_syntax(self) -> EnclosedSyntaxDefinition:
        return self._definitions[SyntaxDefinitionRegistryKey.TOGGLEABLE_HEADING]

    def _create_media_syntax(
        self, media_type: str, url_pattern: str | None = None
    ) -> EnclosedSyntaxDefinition:
        url_pattern = url_pattern or "[^)]+"
        return EnclosedSyntaxDefinition(
            start_delimiter=f"[{media_type}](",
            end_delimiter=")",
            regex_pattern=re.compile(
                rf"(?<!\!)\[{re.escape(media_type)}\]\(({url_pattern})\)"
            ),
            end_regex_pattern=re.compile(r"\)"),
        )

    def _create_url_media_syntax(self, media_type: str) -> EnclosedSyntaxDefinition:
        return EnclosedSyntaxDefinition(
            start_delimiter=f"[{media_type}](",
            end_delimiter=")",
            regex_pattern=re.compile(
                rf"(?<!\!)\[{re.escape(media_type)}\]\((https?://[^\s)]+)\)"
            ),
            end_regex_pattern=re.compile(r"\)"),
        )

    def _register_defaults(self) -> None:
        # Media elements (enclosed)
        self._register_audio_syntax()
        self._register_video_syntax()
        self._register_image_syntax()
        self._register_file_syntax()
        self._register_pdf_syntax()
        self._register_bookmark_syntax()
        self._register_embed_syntax()

        # Lists (simple)
        self._register_bulleted_list_syntax()
        self._register_numbered_list_syntax()
        self._register_todo_syntax()
        self._register_todo_done_syntax()

        # Block containers (enclosed)
        self._register_toggle_syntax()
        self._register_toggleable_heading_syntax()
        self._register_callout_syntax()
        self._register_code_syntax()
        self._register_column_list_syntax()
        self._register_column_syntax()
        self._register_equation_syntax()

        # Text blocks (simple)
        self._register_quote_syntax()
        self._register_heading_syntax()
        self._register_divider_syntax()
        self._register_breadcrumb_syntax()
        self._register_table_of_contents_syntax()
        self._register_table_syntax()
        self._register_table_row_syntax()
        self._register_caption_syntax()
        self._register_space_syntax()

    # Registration methods - SimpleSyntaxDefinition
    def _register_breadcrumb_syntax(self) -> None:
        definition = SimpleSyntaxDefinition(
            start_delimiter="[breadcrumb]",
            regex_pattern=re.compile(r"^\[breadcrumb\]\s*$", re.IGNORECASE),
        )
        self._definitions[SyntaxDefinitionRegistryKey.BREADCRUMB] = definition

    def _register_bulleted_list_syntax(self) -> None:
        definition = SimpleSyntaxDefinition(
            start_delimiter="- ",
            regex_pattern=re.compile(r"^(\s*)-\s+(?!\[[ xX]\])(.+)$"),
        )
        self._definitions[SyntaxDefinitionRegistryKey.BULLETED_LIST] = definition

    def _register_divider_syntax(self) -> None:
        definition = SimpleSyntaxDefinition(
            start_delimiter="---",
            regex_pattern=re.compile(r"^\s*-{3,}\s*$"),
        )
        self._definitions[SyntaxDefinitionRegistryKey.DIVIDER] = definition

    def _register_numbered_list_syntax(self) -> None:
        definition = SimpleSyntaxDefinition(
            start_delimiter="1. ",
            regex_pattern=re.compile(r"^(\s*)(\d+)\.\s+(.+)$"),
        )
        self._definitions[SyntaxDefinitionRegistryKey.NUMBERED_LIST] = definition

    def _register_quote_syntax(self) -> None:
        definition = SimpleSyntaxDefinition(
            start_delimiter="> ",
            regex_pattern=re.compile(r"^>(?!>)\s*(.+)$"),
        )
        self._definitions[SyntaxDefinitionRegistryKey.QUOTE] = definition

    def _register_table_syntax(self) -> None:
        delimiter = self._markdown_grammar.table_delimiter
        definition = SimpleSyntaxDefinition(
            start_delimiter=delimiter,
            regex_pattern=re.compile(
                rf"^\s*{re.escape(delimiter)}(.+){re.escape(delimiter)}\s*$"
            ),
        )
        self._definitions[SyntaxDefinitionRegistryKey.TABLE] = definition

    def _register_table_row_syntax(self) -> None:
        delimiter = self._markdown_grammar.table_delimiter
        definition = SimpleSyntaxDefinition(
            start_delimiter=delimiter,
            regex_pattern=re.compile(
                rf"^\s*{re.escape(delimiter)}([\s\-:|]+){re.escape(delimiter)}\s*$"
            ),
        )
        self._definitions[SyntaxDefinitionRegistryKey.TABLE_ROW] = definition

    def _register_table_of_contents_syntax(self) -> None:
        definition = SimpleSyntaxDefinition(
            start_delimiter="[toc]",
            regex_pattern=re.compile(r"^\[toc\]$", re.IGNORECASE),
        )
        self._definitions[SyntaxDefinitionRegistryKey.TABLE_OF_CONTENTS] = definition

    def _register_todo_syntax(self) -> None:
        definition = SimpleSyntaxDefinition(
            start_delimiter="- [ ]",
            regex_pattern=re.compile(r"^\s*-\s+\[ \]\s+(.+)$"),
        )
        self._definitions[SyntaxDefinitionRegistryKey.TO_DO] = definition

    def _register_todo_done_syntax(self) -> None:
        definition = SimpleSyntaxDefinition(
            start_delimiter="- [x]",
            regex_pattern=re.compile(r"^\s*-\s+\[x\]\s+(.+)$", re.IGNORECASE),
        )
        self._definitions[SyntaxDefinitionRegistryKey.TO_DO_DONE] = definition

    def _register_caption_syntax(self) -> None:
        definition = SimpleSyntaxDefinition(
            start_delimiter="[caption]",
            regex_pattern=re.compile(r"^\[caption\]\s+(\S.*)$"),
        )
        self._definitions[SyntaxDefinitionRegistryKey.CAPTION] = definition

    def _register_space_syntax(self) -> None:
        definition = SimpleSyntaxDefinition(
            start_delimiter="[space]",
            regex_pattern=re.compile(r"^\[space\]\s*$"),
        )
        self._definitions[SyntaxDefinitionRegistryKey.SPACE] = definition

    def _register_heading_syntax(self) -> None:
        definition = SimpleSyntaxDefinition(
            start_delimiter="#",
            regex_pattern=re.compile(r"^(#{1,3})[ \t]+(.+)$"),
        )
        self._definitions[SyntaxDefinitionRegistryKey.HEADING] = definition

    # Registration methods - EnclosedSyntaxDefinition
    def _register_audio_syntax(self) -> None:
        self._definitions[SyntaxDefinitionRegistryKey.AUDIO] = (
            self._create_media_syntax("audio")
        )

    def _register_video_syntax(self) -> None:
        self._definitions[SyntaxDefinitionRegistryKey.VIDEO] = (
            self._create_media_syntax("video")
        )

    def _register_image_syntax(self) -> None:
        self._definitions[SyntaxDefinitionRegistryKey.IMAGE] = (
            self._create_media_syntax("image")
        )

    def _register_file_syntax(self) -> None:
        self._definitions[SyntaxDefinitionRegistryKey.FILE] = self._create_media_syntax(
            "file"
        )

    def _register_pdf_syntax(self) -> None:
        self._definitions[SyntaxDefinitionRegistryKey.PDF] = self._create_media_syntax(
            "pdf"
        )

    def _register_bookmark_syntax(self) -> None:
        self._definitions[SyntaxDefinitionRegistryKey.BOOKMARK] = (
            self._create_url_media_syntax("bookmark")
        )

    def _register_embed_syntax(self) -> None:
        self._definitions[SyntaxDefinitionRegistryKey.EMBED] = (
            self._create_url_media_syntax("embed")
        )

    def _register_callout_syntax(self) -> None:
        definition = EnclosedSyntaxDefinition(
            start_delimiter="[callout]",
            end_delimiter=")",
            regex_pattern=re.compile(
                r'\[callout\](?:\(([^")]+?)(?:\s+"([^"]+)")?\)|(?:\s+([^"\n]+?)(?:\s+"([^"]+)")?)(?:\n|$))'
            ),
            end_regex_pattern=re.compile(r"\)"),
        )
        self._definitions[SyntaxDefinitionRegistryKey.CALLOUT] = definition

    def _register_code_syntax(self) -> None:
        code_delimiter = "```"
        definition = EnclosedSyntaxDefinition(
            start_delimiter=code_delimiter,
            end_delimiter=code_delimiter,
            regex_pattern=re.compile("^" + re.escape(code_delimiter) + r"(\w*)\s*$"),
            end_regex_pattern=re.compile("^" + re.escape(code_delimiter) + r"\s*$"),
        )
        self._definitions[SyntaxDefinitionRegistryKey.CODE] = definition

    def _register_column_syntax(self) -> None:
        delimiter = self._markdown_grammar.column_delimiter
        definition = EnclosedSyntaxDefinition(
            start_delimiter=f"{delimiter} column",
            end_delimiter=delimiter,
            regex_pattern=re.compile(
                rf"^{re.escape(delimiter)}\s*column(?:\s+(0?\.\d+|1(?:\.0?)?))??\s*$",
                re.IGNORECASE | re.MULTILINE,
            ),
            end_regex_pattern=re.compile(rf"^{re.escape(delimiter)}\s*$", re.MULTILINE),
        )
        self._definitions[SyntaxDefinitionRegistryKey.COLUMN] = definition

    def _register_column_list_syntax(self) -> None:
        delimiter = self._markdown_grammar.column_delimiter
        definition = EnclosedSyntaxDefinition(
            start_delimiter=f"{delimiter} columns",
            end_delimiter=delimiter,
            regex_pattern=re.compile(
                rf"^{re.escape(delimiter)}\s*columns?\s*$", re.IGNORECASE
            ),
            end_regex_pattern=re.compile(rf"^{re.escape(delimiter)}\s*$"),
        )
        self._definitions[SyntaxDefinitionRegistryKey.COLUMN_LIST] = definition

    def _register_equation_syntax(self) -> None:
        definition = EnclosedSyntaxDefinition(
            start_delimiter="$$",
            end_delimiter="$$",
            regex_pattern=re.compile(r"^\$\$\s*$"),
            end_regex_pattern=re.compile(r"^\$\$\s*$"),
        )
        self._definitions[SyntaxDefinitionRegistryKey.EQUATION] = definition

    def _register_toggle_syntax(self) -> None:
        delimiter = self._markdown_grammar.toggle_delimiter
        definition = EnclosedSyntaxDefinition(
            start_delimiter=delimiter,
            end_delimiter=delimiter,
            regex_pattern=re.compile(rf"^{re.escape(delimiter)}\s+(.+)$"),
            end_regex_pattern=re.compile(rf"^{re.escape(delimiter)}\s*$"),
        )
        self._definitions[SyntaxDefinitionRegistryKey.TOGGLE] = definition

    def _register_toggleable_heading_syntax(self) -> None:
        delimiter = self._markdown_grammar.toggle_delimiter
        escaped_delimiter = re.escape(delimiter)
        definition = EnclosedSyntaxDefinition(
            start_delimiter=f"{delimiter} #",
            end_delimiter=delimiter,
            regex_pattern=re.compile(
                rf"^{escaped_delimiter}\s*(?P<level>#{{1,3}})(?!#)\s*(.+)$",
                re.IGNORECASE,
            ),
            end_regex_pattern=re.compile(rf"^{escaped_delimiter}\s*$"),
        )
        self._definitions[SyntaxDefinitionRegistryKey.TOGGLEABLE_HEADING] = definition



import asyncio
from collections.abc import Callable

from notionary.blocks.client import NotionBlockHttpClient
from notionary.blocks.schemas import Block
from notionary.page.content.markdown.builder import MarkdownBuilder
from notionary.page.content.parser.service import MarkdownToNotionConverter
from notionary.page.content.renderer.service import NotionToMarkdownConverter
from notionary.utils.decorators import async_retry, time_execution_async
from notionary.utils.mixins.logging import LoggingMixin


class PageContentService(LoggingMixin):
    def __init__(
        self,
        page_id: str,
        block_client: NotionBlockHttpClient,
        markdown_converter: MarkdownToNotionConverter,
        notion_to_markdown_converter: NotionToMarkdownConverter,
    ) -> None:
        self._page_id = page_id
        self._block_client = block_client
        self._markdown_converter = markdown_converter
        self._notion_to_markdown_converter = notion_to_markdown_converter

    @time_execution_async()
    async def get_as_markdown(self) -> str:
        blocks = await self._block_client.get_block_tree(parent_block_id=self._page_id)
        return await self._notion_to_markdown_converter.convert(blocks=blocks)

    @time_execution_async()
    async def clear(self) -> None:
        children_response = await self._block_client.get_block_children(
            block_id=self._page_id
        )

        if not children_response or not children_response.results:
            self.logger.debug("No blocks to delete for page: %s", self._page_id)
            return

        await asyncio.gather(
            *[self._delete_single_block(block) for block in children_response.results]
        )

    @async_retry(max_retries=10, initial_delay=0.2, backoff_factor=1.5)
    async def _delete_single_block(self, block: Block) -> None:
        self.logger.debug("Deleting block: %s", block.id)
        await self._block_client.delete_block(block.id)

    @time_execution_async()
    async def append_markdown(
        self, content: str | Callable[[MarkdownBuilder], MarkdownBuilder]
    ) -> None:
        markdown = self._extract_markdown(content)
        if not markdown:
            self.logger.debug(
                "No markdown content to append for page: %s", self._page_id
            )
            return

        blocks = await self._markdown_converter.convert(markdown)
        await self._append_blocks(blocks)

    def _extract_markdown(
        self, content: str | Callable[[MarkdownBuilder], MarkdownBuilder]
    ) -> str:
        if isinstance(content, str):
            return content

        if callable(content):
            builder = MarkdownBuilder()
            content(builder)
            return builder.build()

        raise ValueError(
            "content must be either a string or a callable that takes a MarkdownBuilder"
        )

    async def _append_blocks(self, blocks: list[Block]) -> None:
        await self._block_client.append_block_children(
            block_id=self._page_id, children=blocks
        )



from typing import ClassVar

from notionary.blocks.rich_text.models import (
    MentionDate,
    MentionType,
    RichText,
    RichTextType,
)
from notionary.blocks.rich_text.name_id_resolver import (
    DatabaseNameIdResolver,
    DataSourceNameIdResolver,
    NameIdResolver,
    PageNameIdResolver,
    PersonNameIdResolver,
)
from notionary.blocks.schemas import BlockColor


class RichTextToMarkdownConverter:
    VALID_COLORS: ClassVar[set[str]] = {color.value for color in BlockColor}

    def __init__(
        self,
        *,
        page_resolver: NameIdResolver | None = None,
        database_resolver: NameIdResolver | None = None,
        data_source_resolver: NameIdResolver | None = None,
        person_resolver: NameIdResolver | None = None,
    ) -> None:
        self.page_resolver = page_resolver or PageNameIdResolver()
        self.database_resolver = database_resolver or DatabaseNameIdResolver()
        self.data_source_resolver = data_source_resolver or DataSourceNameIdResolver()
        self.person_resolver = person_resolver or PersonNameIdResolver()

    async def to_markdown(self, rich_text: list[RichText]) -> str:
        if not rich_text:
            return ""

        parts: list[str] = []

        for rich_obj in rich_text:
            formatted_text = await self._convert_rich_text_to_markdown(rich_obj)
            parts.append(formatted_text)

        return "".join(parts)

    async def _convert_rich_text_to_markdown(self, obj: RichText) -> str:
        if obj.type == RichTextType.EQUATION and obj.equation:
            return f"${obj.equation.expression}$"

        if obj.type == RichTextType.MENTION:
            mention_markdown = await self._extract_mention_markdown(obj)
            if mention_markdown:
                return mention_markdown

        content = obj.plain_text or (obj.text.content if obj.text else "")
        return self._apply_text_formatting_to_content(obj, content)

    async def _extract_mention_markdown(self, obj: RichText) -> str | None:
        if not obj.mention:
            return None

        mention = obj.mention

        if mention.type == MentionType.PAGE and mention.page:
            return await self._extract_page_mention_markdown(mention.page.id)

        elif mention.type == MentionType.DATABASE and mention.database:
            return await self._extract_database_mention_markdown(mention.database.id)

        elif mention.type == MentionType.DATASOURCE and mention.data_source:
            return await self._extract_data_source_mention_markdown(
                mention.data_source.id
            )

        elif mention.type == MentionType.USER and mention.user:
            return await self._extract_user_mention_markdown(mention.user.id)

        elif mention.type == MentionType.DATE and mention.date:
            return self._extract_date_mention_markdown(mention.date)

        return None

    async def _extract_page_mention_markdown(self, page_id: str) -> str:
        page_name = await self.page_resolver.resolve_id_to_name(page_id)
        return f"@page[{page_name or page_id}]"

    async def _extract_database_mention_markdown(self, database_id: str) -> str:
        database_name = await self.database_resolver.resolve_id_to_name(database_id)
        return f"@database[{database_name or database_id}]"

    async def _extract_data_source_mention_markdown(self, data_source_id: str) -> str:
        data_source_name = await self.data_source_resolver.resolve_id_to_name(
            data_source_id
        )
        return f"@datasource[{data_source_name or data_source_id}]"

    async def _extract_user_mention_markdown(self, user_id: str) -> str:
        user_name = await self.person_resolver.resolve_id_to_name(user_id)
        return f"@user[{user_name or user_id}]"

    def _extract_date_mention_markdown(self, date_mention: MentionDate) -> str:
        date_range = date_mention.start
        if date_mention.end:
            date_range += f"–{date_mention.end}"
        return f"@date[{date_range}]"

    def _apply_text_formatting_to_content(self, obj: RichText, content: str) -> str:
        if obj.text and obj.text.link:
            content = f"[{content}]({obj.text.link.url})"

        if not obj.annotations:
            return content

        annotations = obj.annotations

        if annotations.code:
            content = f"`{content}`"
        if annotations.strikethrough:
            content = f"~~{content}~~"
        if annotations.underline:
            content = f"__{content}__"
        if annotations.italic:
            content = f"*{content}*"
        if annotations.bold:
            content = f"**{content}**"

        if (
            annotations.color != BlockColor.DEFAULT
            and annotations.color in self.VALID_COLORS
        ):
            content = f"({annotations.color}:{content})"

        return content


async def convert_rich_text_to_markdown(
    rich_text: list[RichText],
    *,
    page_resolver: NameIdResolver | None = None,
    database_resolver: NameIdResolver | None = None,
    data_source_resolver: NameIdResolver | None = None,
    person_resolver: NameIdResolver | None = None,
) -> str:
    converter = RichTextToMarkdownConverter(
        page_resolver=page_resolver,
        database_resolver=database_resolver,
        data_source_resolver=data_source_resolver,
        person_resolver=person_resolver,
    )
    return await converter.to_markdown(rich_text)



from typing import override

from notionary.blocks.schemas import (
    Block,
    BlockType,
    ExternalFileWithCaption,
    NotionHostedFileWithCaption,
)
from notionary.page.content.renderer.renderers.file_like_block import (
    FileLikeBlockRenderer,
)
from notionary.page.content.syntax.definition import EnclosedSyntaxDefinition


class FileRenderer(FileLikeBlockRenderer):
    @override
    def _can_handle(self, block: Block) -> bool:
        return block.type == BlockType.FILE

    @override
    def _get_syntax(self) -> EnclosedSyntaxDefinition:
        return self._syntax_registry.get_file_syntax()

    @override
    def _get_file_data(
        self, block: Block
    ) -> ExternalFileWithCaption | NotionHostedFileWithCaption | None:
        return block.file

    def _extract_file_name(self, block: Block) -> str:
        if not block.file:
            return ""

        if block.file.name:
            return block.file.name or ""

        return ""



from collections.abc import AsyncGenerator

import httpx

from notionary.file_upload.query.models import FileUploadQuery
from notionary.file_upload.schemas import (
    FileUploadCompleteRequest,
    FileUploadCreateRequest,
    FileUploadListResponse,
    FileUploadResponse,
    UploadMode,
)
from notionary.http.client import NotionHttpClient
from notionary.utils.pagination import (
    PaginatedResponse,
    paginate_notion_api,
    paginate_notion_api_generator,
)


class FileUploadHttpClient(NotionHttpClient):
    async def create_single_part_upload(
        self,
        filename: str,
        content_type: str | None = None,
    ) -> FileUploadResponse:
        return await self._create_upload(
            filename=filename,
            content_type=content_type,
            mode=UploadMode.SINGLE_PART,
            number_of_parts=None,
        )

    async def create_multi_part_upload(
        self,
        filename: str,
        number_of_parts: int,
        content_type: str | None = None,
    ) -> FileUploadResponse:
        return await self._create_upload(
            filename=filename,
            content_type=content_type,
            mode=UploadMode.MULTI_PART,
            number_of_parts=number_of_parts,
        )

    async def send_file_content(
        self,
        file_upload_id: str,
        file_content: bytes,
        filename: str,
        part_number: int | None = None,
    ) -> FileUploadResponse:
        await self._ensure_initialized()

        url = self._build_send_url(file_upload_id)
        files = {"file": (filename, file_content)}
        data = self._build_part_number_data(part_number)

        response = await self._send_multipart_request(url, files=files, data=data)
        return FileUploadResponse.model_validate(response.json())

    async def complete_upload(self, file_upload_id: str) -> FileUploadResponse:
        request = FileUploadCompleteRequest()
        response = await self.post(
            f"file_uploads/{file_upload_id}/complete",
            data=request.model_dump(),
        )
        return FileUploadResponse.model_validate(response)

    async def get_file_upload(self, file_upload_id: str) -> FileUploadResponse:
        response = await self.get(f"file_uploads/{file_upload_id}")
        return FileUploadResponse.model_validate(response)

    async def list_file_uploads(
        self,
        query: FileUploadQuery | None = None,
    ) -> list[FileUploadResponse]:
        query = query or FileUploadQuery()
        return await paginate_notion_api(
            lambda **kwargs: self._fetch_file_uploads_page(query=query, **kwargs),
            total_results_limit=query.total_results_limit,
        )

    async def list_file_uploads_stream(
        self,
        query: FileUploadQuery | None = None,
    ) -> AsyncGenerator[FileUploadResponse]:
        query = query or FileUploadQuery()
        async for upload in paginate_notion_api_generator(
            lambda **kwargs: self._fetch_file_uploads_page(query=query, **kwargs),
            total_results_limit=query.total_results_limit,
        ):
            yield upload

    async def _create_upload(
        self,
        filename: str,
        mode: UploadMode,
        content_type: str | None,
        number_of_parts: int | None,
    ) -> FileUploadResponse:
        request = FileUploadCreateRequest(
            filename=filename,
            content_type=content_type,
            mode=mode,
            number_of_parts=number_of_parts,
        )
        response = await self.post("file_uploads", data=request.model_dump())
        return FileUploadResponse.model_validate(response)

    async def _send_multipart_request(
        self,
        url: str,
        files: dict,
        data: dict | None = None,
    ) -> httpx.Response:
        headers = self._build_multipart_headers()

        async with httpx.AsyncClient(headers=headers, timeout=self.timeout) as client:
            response = await client.post(url, files=files, data=data)

        response.raise_for_status()
        return response

    async def _fetch_file_uploads_page(
        self,
        query: FileUploadQuery,
        start_cursor: str | None = None,
        **kwargs,
    ) -> PaginatedResponse:
        params = query.model_dump(exclude_none=True)
        params["page_size"] = min(query.page_size_limit or 100, 100)

        if start_cursor:
            params["start_cursor"] = start_cursor

        response = await self.get("file_uploads", params=params)
        parsed = FileUploadListResponse.model_validate(response)

        return PaginatedResponse(
            results=parsed.results,
            has_more=parsed.has_more,
            next_cursor=parsed.next_cursor,
        )

    def _build_send_url(self, file_upload_id: str) -> str:
        return f"{self.BASE_URL}/file_uploads/{file_upload_id}/send"

    def _build_part_number_data(self, part_number: int | None) -> dict | None:
        if part_number is not None:
            return {"part_number": str(part_number)}
        return None

    def _build_multipart_headers(self) -> dict:
        return {
            "Authorization": f"Bearer {self.token}",
            "Notion-Version": self.NOTION_VERSION,
        }



from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, override

from notionary.exceptions.file_upload import FileSizeException
from notionary.file_upload.validation.port import FileUploadValidator

if TYPE_CHECKING:
    from notionary.user import BotUser


class FileUploadLimitValidator(FileUploadValidator):
    def __init__(self, filename: str | Path, file_size_bytes: int) -> None:
        self.filename = Path(filename).name if isinstance(filename, Path) else filename
        self.file_size_bytes = file_size_bytes

    @override
    async def validate(self, integration: BotUser | None = None) -> None:
        from notionary.user import BotUser

        integration = integration or await BotUser.from_current_integration()

        max_file_size_in_bytes = integration.workspace_file_upload_limit_in_bytes

        if self.file_size_bytes > max_file_size_in_bytes:
            raise FileSizeException(
                filename=self.filename,
                file_size_bytes=self.file_size_bytes,
                max_size_bytes=max_file_size_in_bytes,
            )



import difflib
from typing import ClassVar

from notionary.exceptions.base import NotionaryException
from notionary.shared.models.parent import ParentType


class PagePropertyNotFoundError(NotionaryException):
    def __init__(
        self,
        page_url: str,
        property_name: str,
        available_properties: list[str] | None = None,
        max_suggestions: int = 3,
        cutoff: float = 0.6,
    ) -> None:
        self.property_name = property_name

        if available_properties:
            self.suggestions = difflib.get_close_matches(
                property_name, available_properties, n=max_suggestions, cutoff=cutoff
            )
        else:
            self.suggestions = []

        message = f"Property '{property_name}' not found."
        if self.suggestions:
            suggestions_str = "', '".join(self.suggestions)
            message += f" Did you mean '{suggestions_str}'?"
        message += f" Please check the page at {page_url} to verify if the property exists and is correctly named."
        super().__init__(message)


class PagePropertyTypeError(NotionaryException):
    def __init__(
        self,
        property_name: str,
        actual_type: str,
    ) -> None:
        message = f"Property '{property_name}' is of type '{actual_type}'. Use the appropriate getter method for this property type."
        super().__init__(message)


class AccessPagePropertyWithoutDataSourceError(NotionaryException):
    _PARENT_DESCRIPTIONS: ClassVar[dict[ParentType, str]] = {
        ParentType.WORKSPACE: "the workspace itself",
        ParentType.PAGE_ID: "another page",
        ParentType.BLOCK_ID: "a block",
        ParentType.DATABASE_ID: "a database",
    }

    def __init__(self, parent_type: ParentType) -> None:
        parent_desc = self._PARENT_DESCRIPTIONS.get(
            parent_type, f"its parent type is '{parent_type}'"
        )
        message = (
            f"Cannot access properties other than title because this page's parent is {parent_desc}. "
            "To use operations like property reading/writing, you need to use a page whose parent is a data source."
        )
        super().__init__(message)



from enum import StrEnum


class PropertyType(StrEnum):
    TITLE = "title"
    RICH_TEXT = "rich_text"
    SELECT = "select"
    MULTI_SELECT = "multi_select"
    STATUS = "status"
    NUMBER = "number"
    DATE = "date"
    CHECKBOX = "checkbox"
    URL = "url"
    EMAIL = "email"
    PHONE_NUMBER = "phone_number"
    PEOPLE = "people"
    CREATED_BY = "created_by"
    LAST_EDITED_BY = "last_edited_by"
    CREATED_TIME = "created_time"
    LAST_EDITED_TIME = "last_edited_time"
    LAST_VISITED_TIME = "last_visited_time"
    FORMULA = "formula"
    ROLLUP = "rollup"
    FILES = "files"
    RELATION = "relation"
    BUTTON = "button"
    LOCATION = "location"
    PLACE = "place"
    VERIFICATION = "verification"
    UNIQUE_ID = "unique_id"
